package uk.co.hcsoftware.commander.bookmarks;


import org.jdom.Element;
import uk.co.hcsoftware.commander.gui.ViewStateTransferData;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.logging.Logger;

/**
 * There are 3 kinds of bookmark:
 * 1. directory: has no payload, just serves to group it's children
 * 2. file: has a list of files, must be a leaf.
 * 3. super: as a view-specification. On launch replaces the visible views with its saved configuration (the view-spec is stored as a jdom Element)
 *
 *
 */
public class Bookmark implements Cloneable{
    private static final Logger log = Logger.getLogger(Bookmark.class.getName());
    private Element el;
    private List<File> files;
    private String name;

    private boolean isPanelBookmark;

    public Bookmark() {
    }

    public Bookmark(String name) {
        this.name = name;
    }

    public boolean isDirectory(){
        return files==null && el==null;
    }

    public Object clone() throws CloneNotSupportedException {
        Bookmark clone= (Bookmark) super.clone();
        if(el!=null){
            //this immutable but jdom will probably complain if an el is multi-parented..
            clone.el=(Element) el.clone();
        }
        //no need to clone files as this is immutable
        return clone;
    }

    /**
     *
     * @return true if this bookmark customizes the whole window 
     */
    public boolean isViewBookmark(){
        return !isDirectory() && el!=null;
    }

    public Element getEl() {
        return el;
    }

    public List<File> getFiles() {
        return files;
    }

    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }

    public String toString() {
        return "bookmark name="+name+" isdir="+isDirectory()+" isSuper="+isViewBookmark();
    }

    public static Bookmark createBookmark(String name, File... files){
        Bookmark b = new Bookmark(name);
        b.files=new ArrayList<File>(Arrays.asList(files));
        return b;
    }
    public static Bookmark createBookmark(String name, Element el){
        Bookmark b = new Bookmark(name);
        b.el=el;
        b.isPanelBookmark=ViewStateTransferData.ELEMENTNAME.equals(b.el.getName());
        return b;
    }
    public Element writeBookmark(){
        if(isDirectory()){
            Element el = new Element("directory");
            el.setAttribute("name",name);
            return el;
        }
        else{
            Element el = new Element("bookmark");
            el.setAttribute("name",name);
            if(isViewBookmark()){
                Element state=new Element("state");
                el.addContent(state);
                state.addContent((Element) this.el.clone());
            }
            else{
                for(File f:files){
                    Element fileEl=new Element("file");
                    fileEl.setAttribute("path",f.getAbsolutePath());
                    el.addContent(fileEl);
                }
            }
            return el;
        }
    }

    public boolean isSinglePanelBookmark() {
        return isPanelBookmark;
    }

    public static Bookmark createBookmark(Element bookmarkElement){
        Bookmark b = new Bookmark(bookmarkElement.getAttributeValue("name"));
        final String name = bookmarkElement.getName();
        if("directory".equals(name)){
            return b;
        }
        else{
            Element stateEl;
            if((stateEl=bookmarkElement.getChild("state"))!=null){
                b.el = (Element) ((Element) stateEl.getChildren().get(0)).clone();
                b.isPanelBookmark=ViewStateTransferData.ELEMENTNAME.equals(b.el.getName());
                log.finest("bname="+b.el.getName());
                return b;
            }
            else{//must be a list of files
                for(Object child:bookmarkElement.getChildren()){
                    if(child instanceof Element){
                        Element el = (Element) child;
                        if("file".equals(el.getName())){
                             b.addFile(new File(el.getAttributeValue("path")));
                        }
                        else if("location".equals(el.getName())){  //old format: <location target="left" url="C:\t\j\bin\foxrun"/>
                            final String urlValue = el.getAttributeValue("url");
                            File f;
                            if(urlValue.startsWith("file:")){
                                try {
                                    URL u=new URL(urlValue);
                                    f= new File(u.getFile());
                                } catch (MalformedURLException e) {
                                    log.warning("couldn't convert converted urltext="+urlValue+" to file ignoring!");
                                    f=null;
                                }
                            }
                            else f=new File(urlValue);
                            b.addFile(f);
                        }
                    }
                }
                return b;
            }
        }
    }

    public void addFile(File file) {
        if(files==null){
            files=new ArrayList<File>();
            el=null;
        }
        files.add(file);
    }
}
