package uk.co.hcsoftware.commander.bookmarks;



import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;
import uk.co.hcsoftware.commander.AppProperties;
import uk.co.hcsoftware.commander.Main;

import javax.swing.event.TreeModelEvent;
import javax.swing.event.TreeModelListener;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreeModel;
import java.beans.XMLDecoder;
import java.io.*;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.prefs.Preferences;


public class BookmarkReader {
    public static final Logger log = Logger.getLogger(BookmarkReader.class.getName());

    //private static final Preferences p = Preferences.userNodeForPackage(BookmarksBar.class);

    private static BookmarkReader instance = new BookmarkReader();

    public static BookmarkReader getInstance(){
        return instance;
    }

    private BookmarkReader() {
    }

    private File bookmarksFile(){
        return new File(AppProperties.getInstance().getPrefsDirectory(),"bookmarks.xml");
    }
    public void saveBookmarks(TreeModel bookmarks) {
        try {
            writeBookmarks(bookmarks,bookmarksFile());
        }  catch (IOException e) {
            log.log(Level.WARNING, "couldn't save bookmarks", e);
        }
    }

    private void writeXml(File toFile, final Document d) throws IOException {
        //create the file if necessary
        if(!toFile.exists()){
            final File parent = toFile.getParentFile();
            log.fine("creating parent="+parent);
            log.fine("result="+parent.mkdirs());
        }
        else log.finer("parent exists");
        final BufferedOutputStream os = new BufferedOutputStream(new FileOutputStream(toFile));
        try {
            XMLOutputter xmlo = new XMLOutputter(Format.getPrettyFormat());
            xmlo.output(d,os);
        } finally {
            os.close();
        }

    }

    /**
     * prior to 20051024 bookmarks were stored in the user's preference tree under the rootpackagename
     * This was changed to store them in a file under the user's home dir for the following reasons:
     * <ol><li>make it easy for the user to copy bookmarks from one machine to another, (e.g. share them with other users) and
     * edit them through means other than the GUI tools provided</li>
     * <li>preparation for anticipated expansion of bookmarks system (they will be more complicated, allow saving of sort,filter prefs etc.)
     * by having a more elaborate storage format than default javabean/xml encoding</li>
     * </ol>
     * The old changeover should be as invisible as possible to the user. E.g. always write in the new format, try to read
     * in the new format - if fails try to read in the old<br/>
     *
     * @return
     * @param addSaveListener
     */
    public DefaultTreeModel readBookmarks(boolean addSaveListener) {
        DefaultTreeModel dtm;
        try {
            //first look for a file and try to read from that
            File f = bookmarksFile();
            if(f.exists()){//read from the boomarks file
                log.fine("reading bookmarks from file="+f);
                dtm= readBookmarks(f);
            }
            else{ //else try to read from prefs (legacy)
                Preferences p = Preferences.userNodeForPackage(Main.class);
                final String bookmarkString = p.get("bookmarks","");
                if (bookmarkString.length()>0) {

                    dtm=  convertList(read(bookmarkString));
                } else { //no bookmarks in a file, or in the prefs tree, so return the 'default' bookmarks, these
                    //are really just samples to make the initial gui look better and give the user and idea of
                    //what to expect
                    DefaultMutableTreeNode root = new DefaultMutableTreeNode();
                    DefaultMutableTreeNode dir = new DefaultMutableTreeNode(new Bookmark("Default bookmarks"));
                    root.add(dir);
                    dir.add(new DefaultMutableTreeNode(Bookmark.createBookmark("Home",new File(System.getProperty("user.home"))),false));
                    dir.add(new DefaultMutableTreeNode(Bookmark.createBookmark("Temp files",new File(System.getProperty("java.io.tmpdir"))),false));
                    DefaultMutableTreeNode dir1 = new DefaultMutableTreeNode(new Bookmark("My bookmarks"));
                    root.add(dir1);
                    dtm= new DefaultTreeModel(root);
                }
            }
        } catch (Exception e) {
            log.log(Level.WARNING,"read bms",e);
            DefaultTreeModel dtm1;
            final List<Bookmark> l = Collections.emptyList();
            dtm1= convertList(l);
            dtm = dtm1;
        }
        if(addSaveListener){
            final DefaultTreeModel dtm1 = dtm;
            dtm.addTreeModelListener(new TreeModelListener() {
                private void save(){
                    log.fine("saving bookmarks in response to treechagne");
                    saveBookmarks(dtm1);
                }
                public void treeNodesChanged(TreeModelEvent e) { save(); }
                public void treeNodesInserted(TreeModelEvent e) { save(); }
                public void treeNodesRemoved(TreeModelEvent e) { save(); }
                public void treeStructureChanged(TreeModelEvent e) { save(); }
            });
        }
        return dtm;
    }

    private DefaultTreeModel convertList(List<Bookmark> l){
        DefaultMutableTreeNode root = new DefaultMutableTreeNode();
        for(Bookmark b:l){
            root.add(new DefaultMutableTreeNode(b,true));
        }
        return new DefaultTreeModel(root);
    }

    private DefaultTreeModel readBookmarks(File fromFile) throws IOException, JDOMException {
        SAXBuilder sb = new SAXBuilder(false);
        final Document d = sb.build(fromFile);

        final DefaultMutableTreeNode root = new DefaultMutableTreeNode(new Bookmark(""));
        final DefaultTreeModel dtm = new DefaultTreeModel(root);
        new Runnable(){
            public void run() {
                addDescendants(d.getRootElement(), root);
            }
            private DefaultMutableTreeNode createNode(Element el, DefaultMutableTreeNode parent) {
                Bookmark b=Bookmark.createBookmark(el);
                final DefaultMutableTreeNode node = new DefaultMutableTreeNode(b);
                parent.add(node);
                return node;
            }
            private void addDescendants(Element root, DefaultMutableTreeNode parent){
                List children=root.getChildren();
                for(Object child:children){
                    if(child instanceof Element){
                        Element el = (Element)child;
                        String qName=el.getName();
                        if("directory".equals(qName)){
                            final DefaultMutableTreeNode node = createNode(el, parent);
                            addDescendants(el,node);//recurse
                        }
                        else if("bookmark".equals(qName)){
                            final DefaultMutableTreeNode dmtn = createNode(el, parent);
                            dmtn.setAllowsChildren(false);
                            //no recurse: bookmarks are always leaves
                        }
                    }
                }
            }
        }.run();
        return dtm;

//            sp.parse(is,new DefaultHandler(){
//                Bookmark b;
//                DefaultMutableTreeNode parent=root;
//
//                public void endElement(String uri, String localName, String qName) {
//                    if("directory".equals(qName)){
//                        parent=(DefaultMutableTreeNode) parent.getParent();
//                        if(parent==null) parent=root;
//                    }
//                }
//
//                public void startElement(String uri, String localName, String qName, Attributes attributes) {
//                    if("directory".equals(qName)){
//                        b=new Bookmark();
//                        final String name = attributes.getValue("name");
//                        b.setName(name);
//                        final DefaultMutableTreeNode node = new DefaultMutableTreeNode(b);
//                        parent.add(node);
//                        parent=node;
//                    }
//                    else if("bookmark".equals(qName)){
//                        b=new Bookmark();
//                        final String name = attributes.getValue("name");
//                        b.setName(name);
//                        final DefaultMutableTreeNode dmtn = new DefaultMutableTreeNode(b);
//                        dmtn.setAllowsChildren(false);
//                        parent.add(dmtn);
//                    }
//                    else if("location".equals(qName)){
//                        String target = attributes.getValue("target");
//                        try {
//                            if("left".equals(target)){
//                                b.addFile(getFile(attributes.getValue("url")));
//                            }
//                            else if("right".equals(target)){
//                                b.addFile(getFile(attributes.getValue("url")));
//                            }
//                        }  catch (NullPointerException e){
//                            log.warning("encountered location el but bookmark was null!");
//                        }
//                    }
//                    else if("view".equals(qName)){
//                        b.setEl()
//                    }
//                }
//            });

    }




    private Document createXmlDocument(String rootTagName)/* throws ParserConfigurationException */{
//        DocumentBuilder db = DocumentBuilderFactory.newInstance().newDocumentBuilder();
//        final Document d = db.newDocument();
//        final Element rootXmlElement = d.createElement(rootTagName);
//        d.appendChild(rootXmlElement);
       return new Document(new Element(rootTagName));
    }
    private void writeBookmarks(final TreeModel bookmarks, File toFile) throws  IOException {
        final Document d = createXmlDocument("bookmarks");
        //recurse down tree writing bookmarks
        Runnable treeWriter = new Runnable() {
            public void run() {
                 DefaultMutableTreeNode rootDmtn = (DefaultMutableTreeNode) bookmarks.getRoot();
                 if(rootDmtn!=null){
                     for (int i=0;i<rootDmtn.getChildCount();i++){
                         DefaultMutableTreeNode child = (DefaultMutableTreeNode) rootDmtn.getChildAt(i);
                         writeNode(d.getRootElement(), child);
                     }
                 }
            }

            private void writeNode(Element parentXmlElement, DefaultMutableTreeNode dmtn){
                //store name, url1, url2
                //url1 and url2 are optional
                //something like <bookmarks><bookmark name="" path=""/><location url="" target="left"/><location url="" target="right"/>
                // <directory name="hello"><bookmark/> <bookmark/></directory>
                // </bookmarks>
                //example <bookmark name="mybookmark" parentpath="mybookmarks"/> <bookmark name="mybookmarks" parentpath=""/>
                //in future maybe: <bookmarks><bookmark name="" path=""/></bookmarks>
                Bookmark b = (Bookmark) dmtn.getUserObject();
                Element bkEl=b.writeBookmark();
                parentXmlElement.addContent(bkEl);
                if(!dmtn.isLeaf()){ //recurse
                    for (int i=0;i<dmtn.getChildCount();i++){
                        DefaultMutableTreeNode child = (DefaultMutableTreeNode) dmtn.getChildAt(i);
                        writeNode(bkEl, child);
                    }
                }
            }

        };
        treeWriter.run();

        writeXml(toFile, d);
    }

    private List<Bookmark> read(String s){
        if (s!=null && s.trim().length()>0) {
            try {
                log.finer("decode s="+s);
                XMLDecoder d = new XMLDecoder(new ByteArrayInputStream(s.getBytes("UTF8")));
                final List<Bookmark> bookmarks = new ArrayList<Bookmark>();
                try {
                    while(true){
                        final uk.co.hcsoftware.commander.Bookmark.BookmarkBean b = (uk.co.hcsoftware.commander.Bookmark.BookmarkBean) d.readObject();
                        if(b!=null){
                            bookmarks.add(b.createBookmark());
                        }
                    }
                } catch (ArrayIndexOutOfBoundsException e) {
                    ;  //not sure how else to determine when there are no more objects
                }
                d.close();
                log.fine("read bookmarks="+bookmarks);
                return bookmarks;
            } catch (UnsupportedEncodingException e) {
                log.log(Level.WARNING,"unsupported encoding utf8",e);
            } catch (ClassCastException e) {
                log.log(Level.WARNING,"ClassCastException",e);
            } catch(Exception e){
                log.log(Level.WARNING,"Exception",e);
            }
        }
        return new ArrayList<Bookmark>();
    }
}
