package uk.co.hcsoftware.commander.gui;

import uk.co.hcsoftware.commander.gui.util.GuiUpdateQueue;

import javax.swing.*;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.util.LinkedList;
import java.util.logging.Level;
import java.util.logging.Logger;


/**
 * Title: CommanderFrameInputBlock
 * Copyright:    Copyright (c) 01-Mar-2006
 *
 * @author N Spurling
 */
public class CommanderFrameInputBlock implements InputBlocker{
    public static final Logger log = Logger.getLogger(CommanderFrameInputBlock.class.getName());

    private CommanderFrame commanderFrame;

    public CommanderFrameInputBlock(CommanderFrame commanderFrame) {
        this.commanderFrame = commanderFrame;
    }

    public boolean isInputBlocked() {
        return inputBlocked;
    }

    //BLOCKING
    private volatile boolean cancellable=false, inputBlocked=false, cancelled=false;
    public boolean wasCancelled() {
        return cancelled;
    }
    /**
     * call from EDT
     *
     * block gui input in cancellable way. If ESC is pressed wasCancelled() will return true (at which point you might want to abort what you're
     * doing and call unblockinput);
     * @param text  initial message adds "(ESC to cancel)" to this
     */
    public void blockInputWithCancel(String text) {
        if(inputBlocked){
            //throw new RuntimeException("Input is already blocked!"){};
            log.log(Level.WARNING,"Input is already blocked, but blockInput called again!",new Throwable());
        }
        cancelled=false;
        cancellable=true;
        commanderFrame.blockInput(getFullMessage(text),new KeyAdapter() {
            public void keyPressed(KeyEvent e) {
                if(KeyEvent.VK_ESCAPE==e.getKeyCode()){
                    if (cancellable && !cancelled) {
                        cancelled=true;
                        setBlockedMessage("Cancelling..");
                    }
                }
            }
        });
        inputBlocked=true;
        log.fine("blocked input text="+text);
    }

    /**
     * call from EDT
     *
     * block GUI input (e.g. keys, mouse) by putting a glasspane over the whole frame
     * blocking is not cancellable (there is no way for user to do anything until unblockinput is called)
     * @param text to display on the glasspane
     * @param pauseBeforeBlockMillis
     */
    public void blockInput(String text, long pauseBeforeBlockMillis) {
        if(inputBlocked){
            //throw new RuntimeException("Input is already blocked!"){};
            log.warning("Input is already blocked, but blockInput called!");
        }
        cancellable=false;
        commanderFrame.blockInput(text);
        inputBlocked=true;
        log.fine("blocked input text="+text);
    }

    /**
     * call from EDT
     */
    public void unblockInput() {
        commanderFrame.unblockInput();
        inputBlocked=false;
        log.fine("unblocked input");
    }
    private GuiUpdateQueue q = new GuiUpdateQueue(200,-1,new LinkedList<Runnable>() {
        public boolean add(Runnable runnable) {
            clear();
            return super.add(runnable);
        }
    });
    public void updateBlockMessage(String message) {
        if(cancellable && cancelled) return;
        final String fullMessage = getFullMessage(message);
        q.postEvent(new Runnable(){
            public void run() {
                commanderFrame.updateBlockedMessage(fullMessage);
            }
        },1);

    }

    private String getFullMessage(String message) {
        final String fullMessage = cancellable?message+"\n\n"+"(ESC to cancel)":message;
        return fullMessage;
    }

    private void setBlockedMessage(final String fullMessage) {
        if(SwingUtilities.isEventDispatchThread()){
            commanderFrame.updateBlockedMessage(fullMessage);
        }
        else{
            SwingUtilities.invokeLater(new Runnable(){
                public void run() {
                    commanderFrame.updateBlockedMessage(fullMessage);
                }
            });
        }
    }

}
