package uk.co.hcsoftware.commander.gui;


import org.jdesktop.swingx.decorator.Filter;
import uk.co.hcsoftware.commander.gui.ctable.CommanderTable;
import uk.co.hcsoftware.commander.gui.ctable.RowHeaderTable;
import uk.co.hcsoftware.commander.gui.ctable.RowNamer;

import javax.swing.*;
import javax.swing.border.Border;
import javax.swing.border.LineBorder;
import java.awt.*;
import java.awt.dnd.DragGestureEvent;
import java.util.LinkedList;
import java.util.logging.Level;
import java.util.logging.Logger;


/**
 * abstracts focus and colourisation for view-panels; subclass and:
 *
 * 1. subclasses must call init() at end of their constructor
 *
 * Title: CommanderViewPanel
 * Copyright:    Copyright (c) 01-Mar-2006
 *
 * @author N Spurling
 *
 *
 */
public abstract class CommanderViewPanel extends JPanel implements Themeable,Commandule,InputBlocker,DndHelper.StartDrag,RowNamer{
    private static final Logger log = Logger.getLogger(CommanderViewPanel.class.getName());
    protected static final int GAP = 6;

    protected final CommanderFrame commanderFrame;
    protected final CommanderFocus cf;

    protected final CommanderTable table;
    protected  RowHeaderTable rowHeaderTable;
    private final InputBlock inputBlockDelegate;
    private JPanel panel;
    protected WaitPanel waitPanel=new WaitPanel();
    protected final StatusUpdater statusUpdater;

    public CommanderViewPanel(LayoutManager layout, CommanderFocus cf, CommanderFrame commanderFrame, CommanderTable ct) {
        super();
        super.setLayout(new OverlayLayout(this));
        panel=new JPanel(layout);
        super.add(waitPanel);
        super.add(panel);

        waitPanel.setVisible(false);

        this.cf=cf;
        this.commanderFrame=commanderFrame;
        setBorder(focusBorder);
        this.table=ct;
        inputBlockDelegate = new InputBlock(waitPanel,panel);
        table.setIntercellSpacing(new Dimension(0,1));
        rowHeaderTable = new RowHeaderTable(table,this);
        statusUpdater = new StatusUpdater(table,cf,this,commanderFrame);

        new AutocompleteJumperJtable(table,new AutocompleteJumperJtable.RowIdentifier() {
            public String getRowId(int viewRow) {
                return getRowName(viewRow);
//                int ro = table.convertRowIndexToModel(viewRow);
//                return getModel().getElement(ro).getName();
            }
        });
    }
    public boolean shouldStartDrag(DragGestureEvent dge) {
        //check drag origin is in model column 1
        Point p=dge.getDragOrigin();
        int viewCol = rowHeaderTable.columnAtPoint(p);
        int viewRow = rowHeaderTable.rowAtPoint(p);
        log.fine("viewCol="+viewCol+" viewRow="+viewRow);
        return viewCol!=-1 && viewRow !=-1;
//        if(viewCol==-1) return false;
//        int modelCol=rowHeaderTable.convertColumnIndexToModel(viewCol);
//        return modelCol==0;
    }

    public Component add(Component comp) {
        return panel.add(comp);
    }
    public void add(Component comp, Object constraints) {
        panel.add(comp, constraints);
    }

    /**
     * must call me as last thing in constructor
     */
    public void init(){
        //manage focus
        cf.addListener(new CommanderFocus.CommanderFocusListener() {
            public void focusChanged(Component lostFocus, Component gainedFocus) {
                log.finer("focus change lost="+lostFocus+" gained="+gainedFocus);
                ensureFocusColoration();
            }
        });

        new ThemeListener(this);
    }

    private Color focusBorderColor,emptyBorderColor;
    public void setColors(Color backColor, Color foreColor, Color selColor) {
        focusBorderColor=selColor;
        emptyBorderColor=backColor;
        table.setBackground(backColor);
        table.setForeground(foreColor);
        table.setSelectionBackground(selColor);
        table.setSelectionForeground(backColor);

        rowHeaderTable.setBackground(backColor);
        rowHeaderTable.setForeground(foreColor);
        

        waitPanel.setColors(foreColor,backColor);

        ensureFocusColoration();
    }
    private void ensureFocusColoration() {
        boolean hasFocus = cf.hasFocus(this);
        lb.setLineColor(hasFocus?focusBorderColor:emptyBorderColor);
    }
    
    private class SettableLineBorder extends LineBorder{
        public SettableLineBorder(Color color, int thickness) {
            super(color, thickness);
        }
        public void setLineColor(Color c){
            lineColor=c;
            repaint();//repaints the whole component - can we paint the border somehow?
        }
    }
    private final int borderWidth =GAP/2;
    private  SettableLineBorder lb = new SettableLineBorder(getBackground(),borderWidth);
    private Border focusBorder=BorderFactory.createCompoundBorder(
            BorderFactory.createEmptyBorder(borderWidth,borderWidth,borderWidth,borderWidth),
            lb
    );




    /**
     * create filter that moves all values in filterColumn that are <1 to top of list
     * //filter that places directories at the top of the list.
        //has to be a filter (cannot extend Sorter) as FilterPipeline checks instanceof Sorter and only allows 1 sorter per column
     * @return
     */
    protected Filter createDirectoryFilter(final int filterColumn) {
        return new Filter(){
            private int[] toPrevious;
            protected void reset() {
                log.fine("reset");
                int inputSize = getInputSize();
                toPrevious = new int[inputSize];
                fromPrevious = new int[inputSize];
                for (int i = 0; i < inputSize; i++) {
                    toPrevious[i] = i;	// reset before sorting
                }
            }
            protected void filter() {
                int[] from = toPrevious.clone();
                int toIndex=0;
                for (int modelIndex:from){
                    //if it's a dir then move it to top of list
                    Long l=(Long) getInputValue(modelIndex,filterColumn);
                    if(l==-1){
                        toPrevious[toIndex++]=modelIndex;
                    }
                }
                for (int modelIndex:from){
                    Long l=(Long) getInputValue(modelIndex,filterColumn);
                    if(l!=-1) toPrevious[toIndex++]=modelIndex;
                }
                // Generate inverse map for implementing convertRowIndexToView();
                for (int i = 0; i < toPrevious.length; i++) {
                    fromPrevious[toPrevious[i]] = i;
                }
            }
            public int getSize() {
                return toPrevious.length;
            }
            protected int mapTowardModel(int row) {
                return toPrevious[row];
            }
            protected void init() {
                toPrevious = new int[0];
            }
        };
    }



    public boolean getFocus(){
        if(inputBlockDelegate.isInputBlocked()){
            log.log(Level.FINEST,"ignoring reqeustfocus as input is blocked");
            return false;
        }
        final boolean b = table.requestFocusInWindow();
        if(!b){
            table.requestFocus();
        }
        return b;
    }
    /**
     * //try and make the dialog the 'active' window in the platform, so we can just start typing without having to click.. which is annoying
     */
    protected void windowToFront() {
        Window w = SwingUtilities.getWindowAncestor(this);
        if(w!=null){
            log.fine("moving to front and requesting focus");
            w.toFront();
            w.requestFocus();
        }
        else log.info("ancestor w was null!");
    }


     private LinkedList<String> previousStates = new LinkedList<String>();
    public void restorePreviousState() throws CannotRestoreSavedStateException {
          if(!previousStates.isEmpty()){
              String state = previousStates.getFirst();
              restoreState(state);
          }
    }
    public void restoreNextState()throws CannotRestoreSavedStateException {
    }






//inputblock delegation
    public boolean wasCancelled() { return inputBlockDelegate.wasCancelled(); }
    public void blockInputWithCancel(String text) { inputBlockDelegate.blockInputWithCancel(text); }
    public void blockInput(String text, long pauseBeforeBlockMillis) {  inputBlockDelegate.blockInput(text, pauseBeforeBlockMillis); }
    public void unblockInput() {
        inputBlockDelegate.unblockInput();
        getFocus();
    }
    public void updateBlockMessage(String status) { inputBlockDelegate.updateBlockMessage(status); }

}
