package uk.co.hcsoftware.commander.gui;

import ns.TaskTimer;
import uk.co.hcsoftware.commander.gui.util.GuiUpdateQueue;

import javax.swing.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.LinkedList;
import java.util.logging.Level;
import java.util.logging.Logger;


/**
 * Title: InputBlock
 * Copyright:    Copyright (c) 29-Mar-2006
 *
 * @author N Spurling
 */
public class InputBlock implements InputBlocker{
    WaitPanel waitPanel;
    private final JPanel panel;
    private static final Logger log = Logger.getLogger(InputBlock.class.getName());
    private TaskTimer tt;
    private ActionListener cancelListener = new ActionListener(){
        public void actionPerformed(ActionEvent e) {
            if (cancellable && !cancelled) {
                cancelled=true;
                setBlockedMessage("Cancelling..");
            }
        }
    };

    public InputBlock(WaitPanel waitPanel, JPanel panel) {
        this.waitPanel = waitPanel;
        this.panel = panel;
    }

    private volatile boolean cancellable=false, inputBlocked=false, cancelled=false;
    public boolean isInputBlocked() {
        return inputBlocked;
    }
    public boolean wasCancelled() {
        return cancelled;
    }
    /**
     * call from EDT
     *
     * block gui input in cancellable way. If ESC is pressed wasCancelled() will return true (at which point you might want to abort what you're
     * doing and call unblockinput);
     * @param text  initial message adds "(ESC to cancel)" to this
     */
    public void blockInputWithCancel(String text) {
        if(inputBlocked){
            log.log(Level.WARNING,"Input is already blocked, but blockInput called again!",new Throwable());
        }
        cancelled=false;
        cancellable=true;
        initTimer();
        synchronized (blockLock) {
            panel.setVisible(false);
            doBlock(text,true);
//            waitPanel.blockInput(text,cancelListener);
            inputBlocked=true;
            blockId++;
        }
        log.fine("blocked input text="+text);
    }
    private int blockId=0;
    /**
     * call from EDT
     *
     * block GUI input (e.g. keys, mouse) by putting a glasspane over the whole frame
     * blocking is not cancellable (there is no way for user to do anything until unblockinput is called)
     * @param text to display on the glasspane
     * @param pauseBeforeBlockMillis if >0 then pause this number of millis before showing teh waitpanel
     * This is useful to prevent 'flashing' if sometimes the task takes a v. short amount of time.
     * It might, however be dangerous in that the user could perform some action in between the task starting and the GUI
     * being physically blocked..
     * The actual block time will be pauseBeforeBlockMillis+timetocreate new thread+time to whenever SwingUtilities.invokeLater() executes it
     */
    public void blockInput(final String text, final long pauseBeforeBlockMillis) {
        if(inputBlocked){
            log.warning("Input is already blocked, but blockInput called!");
        }
        cancellable=false;
        log.fine("block input called");
        initTimer();
        synchronized (blockLock) {
            blockId++;
            final int id=blockId;
            if (pauseBeforeBlockMillis>0) {
                new Thread(){
                    private boolean inGui=false;
                    public void run() {
                        if(!inGui) {
                            try {
                                sleep(pauseBeforeBlockMillis);
                            } catch (InterruptedException e) {
                                log.log(Level.WARNING, "exception", e);
                            }
                            inGui=true;
                            SwingUtilities.invokeLater(this);
                        }
                        else{
                            synchronized (blockLock) {
                                if (id==blockId && inputBlocked) {
                                    doBlock(text, false);
                                }
                                else log.fine("ignoring block as re-blocked or inputBlock is false");
                            }
                        }

                    }
                }.start();
            } else {
                doBlock(text, false);
            }
            inputBlocked=true;

        }
        log.fine("blocked input text="+text);
    }

    private void initTimer() {
        tt = new TaskTimer();
        blockedAfter="";
    }

    private void doBlock(final String text, boolean cancellable) {
        blockedAfter=tt.timeTakenSoFar();
        if (!cancellable) {
            waitPanel.blockInput(text);
        } else {
            waitPanel.blockInput(text,cancelListener);
        }
        panel.setVisible(false);
    }

    private String blockedAfter;
    /**
     * call from EDT
     */
    public void unblockInput() {
        String unblockedAfter;
        synchronized (blockLock) {
            unblockedAfter=tt.timeTakenSoFar();
            panel.setVisible(true);
            waitPanel.unblockInput();
            inputBlocked=false;
        }
        log.fine("unblocked input: blockedAfter="+blockedAfter+" unblockedAfter="+unblockedAfter);
    }
    private final Object blockLock=new Object();
    private GuiUpdateQueue q = new GuiUpdateQueue(200,-1,new LinkedList<Runnable>() {
        public boolean add(Runnable runnable) {
            clear();
            return super.add(runnable);
        }
    });
    public void updateBlockMessage(String message) {
        if(cancellable && cancelled) return;
        final String fullMessage = message;
        q.postEvent(new Runnable(){
            public void run() {
                waitPanel.updateBlockedMessage(fullMessage);
            }
        },1);
    }

    private void setBlockedMessage(final String fullMessage) {
        if(SwingUtilities.isEventDispatchThread()){
            waitPanel.updateBlockedMessage(fullMessage);
        }
        else{
            SwingUtilities.invokeLater(new Runnable(){
                public void run() {
                    waitPanel.updateBlockedMessage(fullMessage);
                }
            });
        }
    }
}
