package uk.co.hcsoftware.commander.gui;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import javax.swing.*;
import javax.swing.tree.TreeModel;
import javax.swing.tree.TreeNode;
import javax.swing.tree.TreePath;
import javax.xml.parsers.*;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import java.io.*;
import java.util.*;
import java.util.Timer;
import java.util.concurrent.Semaphore;
import java.util.logging.Level;
import java.util.logging.Logger;


/**
 * stores (and restores) the expansion state of a JTree to/from a file
 *
 * Assumes all nodes in the tree (including the root) implement TreeNode
 *
 * saves to a file at present, as this is what I want it for, but should be quite easy to change this to produce
 * XML or save somewhere else instead..
 *
 * Title: TreeExpansionStateSaver
 * Copyright:    Copyright (c) 17-Jan-2006
 *
 * @author spurlinn
 */
public class TreeExpansionStateSaver {
    public static final Logger log = Logger.getLogger(TreeExpansionStateSaver.class.getName());
    private boolean saveInNewThread;
    private final File saveFile;
    private Timer t;

    public TreeExpansionStateSaver(File saveFile) {
        this(saveFile, false);
    }
    public boolean saveFileExists(){
        return saveFile.exists();
    }

    /**
     *
     * @param saveFile
     * @param saveInNewThread if true, all saveTreeState(..) calls will return as soon as the expanded paths have been identified
     * Writing the expanded paths to disk will occur in a separate thread. There is a single thread to do the writing
     * , multiple calls get q'd up and executed serially. In fact, if there is more than one operation in the q when the
     * worker thread is ready to save, then only the final operation will be executed (the others are discarded)
     * since the effect is the same as if they were executed serially
     *
     *
     */
    public TreeExpansionStateSaver(File saveFile, boolean saveInNewThread) {
        this.saveFile = saveFile;
        setSaveInNewThread(saveInNewThread);
    }
    /**
     * see constructor javadocs
     * @param saveInNewThread
     */
    public void setSaveInNewThread(boolean saveInNewThread) {
        this.saveInNewThread = saveInNewThread;
        if(saveInNewThread) t=new Timer(true);
    }

    public void saveTreeState(final JTree jt) {
        final List<TreePath> expandedPaths = new ArrayList<TreePath>();
        //recurse down tree looking for xpanded nodes
        new Runnable() {
            public void run() {
                final TreePath rootPath = new TreePath(jt.getModel().getRoot());
                nodeExpanded(rootPath);
            }
            /**
             * if this path is expanded, recurses to all children, which add themselves to the expandedPaths
             * list if they are the last component in a complete expanded path
             * @param tp
             * @return true if this path is expanded
             */
            private boolean nodeExpanded(TreePath tp){
                final boolean expanded = jt.isExpanded(tp);
                if(expanded){
                    TreeNode tn = (TreeNode) tp.getLastPathComponent();

                    boolean noExpandedChildren=true;
                    final Enumeration children = tn.children();
                    while (children.hasMoreElements()) {
                        TreeNode child = (TreeNode) children.nextElement();
                        if(!child.getAllowsChildren()){
                            continue;
                        }
                        TreePath childPath=tp.pathByAddingChild(child);
                        if(nodeExpanded(childPath)){ //recurse
                            noExpandedChildren=false;
                        }
                    }
                    if(noExpandedChildren){
                        expandedPaths.add(tp);
                    }
                }
                return expanded;
            }
        }.run();

        if (expandedPaths.size()>0) {
            saveExpansionState(expandedPaths);
        }
        else log.fine("0 expanded nodes");
    }
    /**
     * The TreeModel must have the same structure as when saveExpansionState was called (or results are undefined)
     * @param jt
     */
    public void restoreTreeState(JTree jt) throws CannotRestoreStateException{
        // read expanded nodes and expand them
        List<TreePath> paths = getExpansionPaths(jt.getModel());
        for (Iterator<TreePath> iterator = paths.iterator(); iterator.hasNext();) {
            TreePath treePath = iterator.next();
            jt.expandPath(treePath);
        }
    }
    public static final class CannotRestoreStateException extends Exception{
        public CannotRestoreStateException() {
            super();
        }

        public CannotRestoreStateException(String message) {
            super(message);
        }

        public CannotRestoreStateException(String message, Throwable cause) {
            super(message, cause);
        }

        public CannotRestoreStateException(Throwable cause) {
            super(cause);
        }
    }



    private List<TreePath> getExpansionPaths(final TreeModel tm) throws CannotRestoreStateException {
        final List<TreePath> expansionPaths = new ArrayList<TreePath>();
        if(saveFile.exists()){//read from the boomarks file
            try {
                SAXParser sp = SAXParserFactory.newInstance().newSAXParser();
                InputStream is = new BufferedInputStream(new FileInputStream(saveFile));
                final TreeNode rootNode;
                try {
                    rootNode = (TreeNode) tm.getRoot();
                } catch (ClassCastException e) {
                    throw new CannotRestoreStateException(e);
                }
                sp.parse(is,new DefaultHandler(){
                    private boolean readingExpandedNodes=false;
                    public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException {
                        if("expandednodes".equals(qName)){
                            readingExpandedNodes=true;
                        }
                        else if(readingExpandedNodes && "node".equals(qName)){
                            try {
                                TreePath tp = decodePath(attributes.getValue("path"),rootNode);
                                if(tp!=null) expansionPaths.add(tp);
                            } catch (CannotRestoreStateException e) {
                                throw new SAXException(e);
                            }
                        }
                    }
                });
            } catch (ParserConfigurationException e) {
                log.log(Level.WARNING, "exception", e);
            } catch (SAXException e) {
                if(e.getException() instanceof CannotRestoreStateException){
                    throw (CannotRestoreStateException)e.getException();
                }
                log.log(Level.WARNING, "exception", e);
            } catch (IOException e) {
                log.log(Level.WARNING, "exception", e);
            }
        }
        return expansionPaths;
    }
    /**
     * only bother passing the final expanded node in the TreePath, as we can get the path via dmtn.getPath() or some such
     * @param expandedNodes
     */
    private void saveExpansionState(final List<TreePath> expandedNodes) {
        if(saveInNewThread){
            //we want the first call to saveEx.. to spawn a new thread that does the save
            //subsequent calls should wait until any other threads have finished writing before saving their state
            //furthermore, if a thread is waiting to save then any further threads should replace the waiting thread
            //(since all saves happen in series, there's no point saving, then saving again.. might as well skip the saving step)

            //implementation: have a mutex plus an id the id increments for each new call, if when a thread
            //wakes up its copy of the id != latest idvalue then a later brother must have arrived while it was sleeping
            //so it can return immediately
            final int id;
            synchronized(executionThreadLock){
                id=++excecutionThreadId;
            }
            t.schedule(new TimerTask() {
                public void run() {
                    if(log.isLoggable(Level.FINER)) log.finer("id="+id+" wait on mutex t="+Thread.currentThread());
                    executionMutex.acquireUninterruptibly();
                    if(log.isLoggable(Level.FINER)) log.finer("id="+id+" got mutex t="+Thread.currentThread());
                    try{
                        //we have the mutex!
                        if(id!=excecutionThreadId){
                            if (log.isLoggable(Level.FINE)){
                                log.fine("ignoring saver="+id+" as a further executor has arrived while it was sleeping");
                            }
                            //someone else arrived while I was sleeping!
                        }
                        else{
                            if (log.isLoggable(Level.FINE))  log.fine("saver="+id+" is saving t="+Thread.currentThread());
                            saveToXml(expandedNodes);
                            if (log.isLoggable(Level.FINE))  log.fine("saver="+id+" completed t="+Thread.currentThread());
                        }
                    } finally {
                        if(log.isLoggable(Level.FINER)) log.finer("id="+id+" release mutex t="+Thread.currentThread());
                        executionMutex.release();
                    }
                }
            },0);
        }
        else saveToXml(expandedNodes);
    }

    private Object executionThreadLock=new Object();
    private Semaphore executionMutex = new Semaphore(1,false);
    private int excecutionThreadId=0;

    private void saveToXml(List<TreePath> expandedNodes) {
        try {
            DocumentBuilder db = DocumentBuilderFactory.newInstance().newDocumentBuilder();
            final Document d1 = db.newDocument();
            final Element rootXmlElement1 = d1.createElement("expandednodes");
            d1.appendChild(rootXmlElement1);
            final Document d = d1;
            Element rootXmlElement=d.getDocumentElement();
            for (Iterator<TreePath> iterator = expandedNodes.iterator(); iterator.hasNext();) {
                TreePath tp = iterator.next();
                Element bkEl=d.createElement("node");
                bkEl.setAttribute("path",encodePath(tp));
                rootXmlElement.appendChild(bkEl);
            }
            writeXml(saveFile, d);
        } catch (ParserConfigurationException e) {
            log.log(Level.WARNING, "couldn't save treeexpansionstate", e);
        } catch (TransformerException e) {
            log.log(Level.WARNING, "couldn't save treeexpansionstat", e);
        } catch (IOException e) {
            log.log(Level.WARNING, "couldn't save treeexpansionstat", e);
        }
    }

    /**
     * path of form 0/3/1/ where the numbers indicate the index of the node wrt its siblings. There is always a trailing /
     * and path always starts with 0, meaning the root
     * all nodes (including root) must implement TreeNode
     * @param tp
     * @return
     */
    private String encodePath(TreePath tp){
        StringBuilder sb = new StringBuilder();
        TreeNode parent=null;
        for(int i=0;i<tp.getPathCount();i++){
            TreeNode node=(TreeNode) tp.getPathComponent(i);
            final int index;
            if(parent==null) index=0;
            else index=parent.getIndex(node);
            sb.append(index+"/");
            parent=node;
        }
        return sb.toString();
    }
    private TreePath decodePath(String coded, TreeNode rootNode) throws CannotRestoreStateException {
        try {
            String[] indices = coded.split("/");
            log.fine("decode path="+Arrays.asList(indices));
            TreeNode[] nodes = new TreeNode[indices.length];
            TreeNode parent=null;
            for (int i = 0; i < indices.length; i++) { //end at length-1 as there is a trailing slash
                int index = Integer.parseInt(indices[i]);
                if(parent==null){
                    nodes[i]=rootNode;
                }
                else nodes[i]=parent.getChildAt(index);
                parent= nodes[i];
            }
            return new TreePath(nodes);
        } catch (ArrayIndexOutOfBoundsException e) {
            log.warning("error decoding path="+coded);
//            throw new CannotRestoreStateException(e);
            return null;
        }
    }

    private void writeXml(File toFile, final Document d) throws TransformerException, IOException {
        Transformer t = TransformerFactory.newInstance().newTransformer();

        //create the file if necessary
        if(!toFile.exists()){
            final File parent = toFile.getParentFile();
            log.fine("creating parent="+parent);
            log.fine("result="+parent.mkdirs());
        }
        else log.finer("parent exists");
        final BufferedOutputStream os = new BufferedOutputStream(new FileOutputStream(toFile));
        final DOMSource source = new DOMSource(d.getDocumentElement());
        t.transform(source,new StreamResult(os));
        os.close();
    }

}
