package uk.co.hcsoftware.commander.gui;

import org.jdom.Element;
import uk.co.hcsoftware.commander.CannotSaveStateException;
import uk.co.hcsoftware.commander.gui.file.DirectoryPanel;
import uk.co.hcsoftware.commander.gui.search.FileSearchPanel;
import uk.co.hcsoftware.commander.gui.util.WeakListener;
import uk.co.hcsoftware.commander.gui.zip.ZipPanel;
import uk.co.hcsoftware.commander.io.ImageProvider;
import uk.co.hcsoftware.commander.prefs.GeneralConfigPanel;

import javax.swing.*;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.prefs.PreferenceChangeEvent;
import java.util.prefs.PreferenceChangeListener;
import java.util.prefs.Preferences;



/**
 * Title: ViewPanel
 *
 * viewpanel is a display panel that has 2 modes:
 *
 * mode1: contains a single 'commandule' (a gui component that displays something such as a directory or a zipfile)
 * mode2: contains 2 ViewPanels, divided by a JSplitPane
 *
 * Copyright:    Copyright (c) 24-Jan-2006
 *
 * @author spurlinn
 */
public class ViewPanel extends JPanel implements PreferenceChangeListener{//is a jpanel+splitpane
    private static final Logger log=Logger.getLogger(ViewPanel.class.getName());
    private Commandule c;
    private ViewPanel left,right,parent;
    private int splitOrientation;
    private JSplitPane jsp;
    private final CommanderFrame cf;
    private JPanel toolbar;
    private ViewPanel(CommanderFrame cf){
        super(new BorderLayout());
        this.cf=cf;
        
        final Preferences viewPrefs=GeneralConfigPanel.showViewControlsPref.getPrefs();
        WeakListener wl = new WeakListener(this,PreferenceChangeListener.class){
            public Object onNullObserver() {
                viewPrefs.removePreferenceChangeListener((PreferenceChangeListener) getProxy());
                return null;
            }
        };
        viewPrefs.addPreferenceChangeListener((PreferenceChangeListener) wl.getProxy());
    }
    public void preferenceChange(PreferenceChangeEvent evt) {
        if(GeneralConfigPanel.showViewControlsPref.isThisPref(evt)){//viewcontrols value has changed
            log.fine("received prefchagned evt="+evt);
            if(!isTwinMode()){//single mode.. is vc panel showing?
                boolean showToolbar=GeneralConfigPanel.showViewControlsPref.getValue();
                final boolean toolbarVisible=toolbar!=null;
                if(showToolbar!=toolbarVisible){
                    //toggle the current state (either add or remove)
                    SwingUtilities.invokeLater(new Runnable() {
                        public void run() {
                            if(toolbarVisible){ //remove
                                remove(toolbar);
                                toolbar=null;
                            }
                            else{//add
                                toolbar=createToolbar();
                                add(toolbar,BorderLayout.EAST);
                            }
                            validate();
                        }
                    });
                }
            }
        }
    }

    public ViewPanel(CommanderFrame cf, Commandule c) {
        this(cf);
        setSingleMode(c);
    }
    public ViewPanel(CommanderFrame cf, int splitOrientation, ViewPanel left, ViewPanel right) {
        this(cf);
        setTwinMode(splitOrientation,left,right);
    }
    private void setSingleMode(Commandule c) {
        removeAll();
        this.c = c;
        this.left=null;
        this.right=null;
        if (GeneralConfigPanel.showViewControlsPref.getValue()) {
            toolbar = createToolbar();
            add(toolbar,BorderLayout.EAST);
        }
        add((Component) c);
    }
    private JButton getButton(Icon icon, String name, String tooltipText){
        JButton jb = new JButton(name, icon);
        jb.setFocusable(false);
        if(tooltipText!=null) jb.setToolTipText(tooltipText);
        return jb;
    }
    private JPanel createToolbar() {
        JPanel jp = new JPanel();//new FlowLayout(FlowLayout.RIGHT,2,2));
        jp.setFocusable(false);
        BoxLayout lt = new BoxLayout(jp, BoxLayout.PAGE_AXIS);
        jp.setLayout(lt);

        jp.setBorder(BorderFactory.createEmptyBorder(1,0,1,1));

        final JButton[] jbs = {
            getButton(ImageProvider.getInstance().getIcon("icons/16/view_top_bottom.png"),"","Split horizontally")//close
            ,getButton(ImageProvider.getInstance().getIcon("icons/16/view_left_right.png"),"","Split vertically")
            ,getButton(ImageProvider.getInstance().getIcon("icons/16/view_right.png"),"","Unsplit this view")//close
            ,getButton(ImageProvider.getInstance().getIcon("icons/16/editdelete.png"/*"icons/16/view_nofullscreen.png"*/),"","Close this view")//close
        };
        ActionListener al = new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                Object src = e.getSource();
                if(jbs[0]==src){//split h
                    if(isTwinMode()){
                        log.warning("splith called but twinmode: ignoring");
                        return;
                    }
                    split(JSplitPane.VERTICAL_SPLIT);
                }
                else if(jbs[1]==src){//split vert
                    if(isTwinMode()){
                        log.warning("splitv called but twinmode: ignoring");
                        return;
                    }
                    split(JSplitPane.HORIZONTAL_SPLIT);
                }
                else if(jbs[2]==src){//unsplit
                    if(isTwinMode()){
                        log.warning("unsplit called but twinmode: ignoring");
                        return;
                    }
                    ViewPanel parent = getParentViewPanel();
                    if (parent!=null) {
                        parent.unsplit(c);
                        c.getFocus();
                    }
                    else log.warning("!twinmode but no parent! ignoring unsplit");
                }
                else if(jbs[3]==src){ //close
                    close();
                }
            }
        };
        Dimension buttonSize=new Dimension(20,15);
        for (int i = jbs.length-1; i >=0; i--) {
            JButton jb = jbs[i];
            jb.addActionListener(al);
            jp.add(jb);
            jb.setPreferredSize(buttonSize);
        }
        return jp;
    }
    public void split(int orientation) {
        if(isTwinMode()){
            log.warning("close called but twinmode: ignoring");
            return;
        }
        final Commandule c=this.c;
        final Commandule newCommandule = c.createClone();
        split(newCommandule,orientation);
        try {
            newCommandule.restoreState(c.saveState());
            newCommandule.getFocus();
        } catch (CannotRestoreSavedStateException e) {
            log.log(Level.WARNING, "exception", e);
        } catch (CannotSaveStateException e) {
            log.log(Level.WARNING, "exception", e);
        }
    }
    public void close(){
        if(isTwinMode()){
            log.warning("close called but twinmode: ignoring");
            return;
        }
        final ViewPanel parentViewPanel = getParentViewPanel();
        ViewPanel sibling = getSiblingHolder();
        if(sibling!=null && parentViewPanel!=null){
            parentViewPanel.set(sibling);
            parentViewPanel.getFocus();
        }
    }

    private void split(Commandule newCommandule, int splitOrientation,boolean validate){
        //split me in half!
        ViewPanel left=new ViewPanel(cf,c);
        ViewPanel right=new ViewPanel(cf,newCommandule);
        setTwinMode(splitOrientation, left, right);
        if (validate) {
            validate();
        }
    }
    public void split(Commandule newCommandule, int splitOrientation){
        split(newCommandule,splitOrientation,true);
    }

    private void setTwinMode(int splitOrientation, ViewPanel left, ViewPanel right) {
        removeAll();
        jsp = new JSplitPane(splitOrientation);
        this.splitOrientation=splitOrientation;
        cf.setupSplitPane(jsp, 0.5f);
        this.left = left;
        this.right=right;
        left.parent=this;
        right.parent=this;
        jsp.add(left);
        jsp.add(right);
        add(jsp);
        c=null;
    }

    public void unsplit(Commandule preserve){
        if (isTwinMode()) {
            Commandule c=preserve==null?getCommandules().get(0):preserve;
            setSingleMode(c);
            validate();
        }
    }

    /**
     * make this th qualitatively identical to replace by copying references to its members
     * @param replace
     */
    public void set(ViewPanel replace){
        if(replace.isTwinMode()){
            setTwinMode(replace.splitOrientation, replace.left, replace.right);
        }
        else {
            setSingleMode(replace.c);
        }
        validate();
    }

    private boolean isTwinMode() {
        return this.c==null;
    }

    public java.util.List<Commandule> getCommandules(){
        java.util.List<Commandule> l = new ArrayList<Commandule>();
        addCommandules(l);
        return l;
    }
    private void addCommandules(java.util.List<Commandule> l){
        if(!isTwinMode()) l.add(c);
        else{
            left.addCommandules(l);
            right.addCommandules(l);
        }
    }
    /**
     *
     * @param commandule
     * @return the tabholder that hosts this commandule (all commandules ahve a tabholder, only some tabholders have a commanduel)
     */
    public ViewPanel findOwner(Commandule commandule) {
        if(!isTwinMode()){
            return c.equals(commandule)?this:null;
        }
        else{
            ViewPanel th=left.findOwner(commandule);
            if(th!=null) return th;
            return right.findOwner(commandule);
        }
    }
    /**
     *
     * @return the 'sibling' ViewPanel in the structure, or null if I am the root
     */
    public ViewPanel getSiblingHolder(){
        ViewPanel parent=getParentViewPanel();
        if(parent!=null){
            if(parent.left==this) return parent.right;
            else if(parent.right==this) return parent.left;
            log.warning("My parent does not recognize me! (returning null)");
        }
        return null;
    }
    public ViewPanel getParentViewPanel() {
         return parent;
    }

    public void getFocus() {
        if(isTwinMode()){
            left.getFocus();
        }
        else{
            c.getFocus();
        }
    }
    /**
     * store this tabholder's state (including the state of all its descendants)
     * @return
     */
    public Element saveState() {
        final Element viewElement;
        if(isTwinMode()){
            viewElement= new Element("view");
            viewElement.setAttribute("dividerlocation",String.valueOf(cf.getDividerLocation(jsp)));
            viewElement.setAttribute("dividerorientation",String.valueOf(splitOrientation));
            viewElement.addContent(left.saveState());
            viewElement.addContent(right.saveState());
        }
        else{
            viewElement=getViewElement(this.c);
        }
        viewElement.setAttribute("leaf",Boolean.toString(!isTwinMode()));
        return viewElement;
    }

    public static Element getViewElement(Commandule c) {
        Element viewElement = new Element("view");
        viewElement.setAttribute("type",getType(c));
        try {
            viewElement.setAttribute("state",c.saveState());
        } catch (CannotSaveStateException e) {
            log.log(Level.WARNING, "CannotSaveStateException", e);
        }
        return viewElement;
    }

    private static ViewPanel createViewPanel(CommanderFrame cf,Element el){
        if ("view".equals(el.getName())) { //is an old-school bookmarked view
            if(Boolean.parseBoolean(el.getAttributeValue("leaf"))){
                Commandule c= getCommandule(cf, el);
                log.fine("creating new commandule tabholder c="+c);
                final ViewPanel th = new ViewPanel(cf,c);
                log.fine("created th="+th);
                return th;
            }
            else{
                List children = el.getChildren();
                ViewPanel left = ViewPanel.createViewPanel(cf,(Element)children.get(0));
                ViewPanel right = ViewPanel.createViewPanel(cf,(Element)children.get(1));
                if(left==null || right==null){
                    log.warning("l or r is null! l="+left+" right="+right);
                }
                final ViewPanel th = new ViewPanel(cf,Integer.parseInt(el.getAttributeValue("dividerorientation")),left,right);

                final String divLoc = el.getAttributeValue("dividerlocation");
                log.fine("divLoc="+divLoc);
                double dl = Double.parseDouble(divLoc);
                if(dl>1) dl=1;
                if(dl<0) dl=0;

                th.jsp.setDividerLocation(dl);
                log.fine("created twin th="+th);
                return th;
            }
        } else { //is a new single-pane bookmark (toplevel el is viewstate)
            Element viewChild=el.getChild("view");
            Commandule c= getCommandule(cf, viewChild);
            log.fine("creating new commandule tabholder c="+c);
            final ViewPanel th = new ViewPanel(cf,c);
            log.fine("created th="+th);
            return th;
        }
    }

    public static Commandule getCommandule(CommanderFrame cf, Element viewElement) {
        Commandule c = getCommandule(cf,viewElement.getAttributeValue("type"));
        try {
            final String stateValue = viewElement.getAttributeValue("state");
            if (stateValue!=null) {
                c.restoreState(stateValue);
            } else {
                log.warning("null state for el="+viewElement);
            }
        } catch (CannotRestoreSavedStateException e) {
            log.log(Level.WARNING, "couldn't restore state", e);
        }
        return c;
    }

    private static Commandule getCommandule(CommanderFrame cf,String type){
        if("ZipPanel".equals(type)){
            return new ZipPanel(cf,cf.getCommanderFocus());
        }
        else if("DirectoryPanel".equals(type)){
            return new DirectoryPanel(cf,cf.getCommanderFocus());
        }
        else if("FileSearchPanel".equals(type)){
            return new FileSearchPanel(cf,cf.getCommanderFocus());
        }
        else{
            log.warning("unrecognized type="+type+" assuming dirpanel");
            return new DirectoryPanel(cf,cf.getCommanderFocus());
        }
    }
    private static String getType(Commandule c) {
        String type;
        if(c instanceof ZipPanel){
            type="ZipPanel";
        }
        else if(c instanceof DirectoryPanel){
            type="DirectoryPanel";
        }
        else if(c instanceof FileSearchPanel){
            type="FileSearchPanel";
        }
        else{
            log.warning("couldn't get type of c="+c+" assuming dirpanel");
            type= "DirectoryPanel";
        }
        return type;
    }
    public void restoreState(Element el){
        final ViewPanel viewPanel = createViewPanel(cf,el);
        log.fine("restoring to th="+viewPanel);
        set(viewPanel);
    }

    public String toString() {
        return "tabholder c="+c+" l="+left+" r="+right;
    }

    /**
     * replace the current commandule with this one
     * @param commandule
     */
    public void setCommandule(Commandule commandule) {
        setSingleMode(commandule);
        validate();
    }
}
