package uk.co.hcsoftware.commander.gui.ctable;

import org.jdesktop.swingx.JXTable;
import org.jdesktop.swingx.decorator.ComponentAdapter;
import org.jdesktop.swingx.decorator.Highlighter;
import org.jdesktop.swingx.decorator.HighlighterPipeline;
import uk.co.hcsoftware.commander.gui.CommanderFocus;

import javax.swing.*;
import java.awt.*;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;


/**
 * utility for gettign selection/mark informtion.
 *
 * Useful if and only if:
 *
 * 1. use MarkableTableModel to store mark information
 *
 * Title: CommanderSelectionUtilities
 * Copyright:    Copyright (c) 20-Mar-2006
 *
 * @author N Spurling
 */
public class CommanderSelectionUtilities<E> {
    public static final Logger log = Logger.getLogger(CommanderSelectionUtilities.class.getName());

    CommanderTable table;
    MarkableTableModel<E> mtm;

    public CommanderSelectionUtilities(final CommanderTable table, final CommanderFocus cf, final Component commanderFocusable) {
        this.table = table;
        this.mtm = (MarkableTableModel<E>) table.getModel();
        log.finer("cteating selutils for table="+table);
        //add a highlighter to draw selected rows, need to use a highlighter so can avoid drawing selected rows if
        //we don't have the commanderfocus. Also paints background if model says row is marked
        Highlighter[] hl = {new Highlighter(){
            private boolean isMarked(ComponentAdapter adapter) {
                final int viewRow = adapter.row;
                final int modelRow = table.convertRowIndexToModel(viewRow);
                log.finest("vrow="+viewRow+" mrow="+modelRow+" ismarked="+mtm.isMarked(modelRow));
                return mtm.isMarked(modelRow);
            }
            protected Color computeBackground(Component renderer, ComponentAdapter adapter) {
                if(!cf.hasFocus(commanderFocusable)){
                    return table.getBackground();
                }
                final boolean selected = adapter.isSelected();
//                final int modelColumn = adapter.viewToModel(adapter.column);
                if(selected /*&& !(modelColumn==0 && marked)*/){ //if model column 0 (mark column) and is marked, then use mark colouration
                    //this is so that if you have selected a row and toggle the mark-status, there is some visual indication that you have changed something
                    return table.getSelectionBackground();
                }
                else if(isMarked(adapter)){
                    final Color selectionBackground = table.getSelectionBackground();
                    Color markColor = selectionBackground.brighter();
                    if(markColor.equals(table.getBackground()) || markColor.equals(selectionBackground)){
                        //if making the colour brighter doesn't have any effect because it's already max brightness
                        //or if it pushes the sel into being the same colour as the background,
                        //then make it darker instead
                        markColor=selectionBackground.darker();
                    }
                    return markColor;
                }
                else return table.getBackground();
            }
            protected Color computeForeground(Component renderer, ComponentAdapter adapter) {
                if(!cf.hasFocus(commanderFocusable)){
                    return table.getForeground();
                }
                if(adapter.isSelected() || isMarked(adapter)) return table.getSelectionForeground();
                return table.getForeground();
            }
        }
        };
        table.setHighlighters(new HighlighterPipeline(hl));
    }

    public boolean isSelectedFiles(boolean includeSelected, boolean includeMarkedFiles){
        if(includeSelected && table.getSelectedRowCount()>0){
            return true;
        }
        if(includeMarkedFiles){
            int count = table.getRowCount();
            for(int ro=0;ro<count;ro++){
                int modelRow=table.convertRowIndexToModel(ro);
                if(mtm.isMarked(modelRow)){
                    return true;
                }
            }
        }
       return false;
    }

    /**
     * return view indices of selected rows
     *
     * @param includeSelected
     * @param includeMarkedFiles
     * @return selected _view_ row indices
     */
    public int[] getSelectedRows(boolean includeSelected, boolean includeMarkedFiles){
        final int rowCount = table.getRowCount();
        ListSelectionModel selectionModel=table.getSelectionModel();
        java.util.List<Integer> l = new ArrayList<Integer>();
        for(int viewRow = 0; viewRow < rowCount; viewRow++) {
            int modelRow=table.convertRowIndexToModel(viewRow);
            if (includeSelected && selectionModel.isSelectedIndex(viewRow)) {
                l.add(viewRow);
            }
            else if(includeMarkedFiles && mtm.isMarked(modelRow)){//(Boolean) table.getValueAt(viewRow, 0)).booleanValue()){
                l.add(viewRow);
            }
        }
        int[] ret=new int[l.size()];
        int j=0;
        for(Integer i:l){
            ret[j++]=i;
        }
        return ret;
    }
    public static int getSelectedRowCount(JXTable table, MarkableTableModel mtm, boolean includeSelected, boolean includeMarkedFiles){
        int count=0;
        final int rowCount = table.getRowCount();
        ListSelectionModel selectionModel=table.getSelectionModel();
        for(int viewRow = 0; viewRow < rowCount; viewRow++) {
            int modelRow=table.convertRowIndexToModel(viewRow);
            if (includeSelected && selectionModel.isSelectedIndex(viewRow)) {
                count++;
            }
            else if(includeMarkedFiles && mtm.isMarked(modelRow)){//(Boolean) table.getValueAt(viewRow, 0)).booleanValue()){
                count++;
            }
        }
       return count;
    }
    public int getSelectedRowCount(boolean includeSelected, boolean includeMarkedFiles){
        return getSelectedRowCount(table,mtm,includeSelected,includeMarkedFiles);
    }

    public List<E> getSelectedElements(boolean includeSelected, boolean includeMarkedFiles){
        List<E> l = new ArrayList<E>();
        final int rowCount = table.getRowCount();
        ListSelectionModel selectionModel=table.getSelectionModel();
        for(int viewRow = 0; viewRow < rowCount; viewRow++) {
            int modelRow=table.convertRowIndexToModel(viewRow);
            if (includeSelected && selectionModel.isSelectedIndex(viewRow)) {
                l.add(mtm.getElement(modelRow));
            }
            else if(includeMarkedFiles && mtm.isMarked(modelRow)){//(Boolean) table.getValueAt(viewRow, 0)).booleanValue()){
                l.add(mtm.getElement(modelRow));
            }
        }
       return l;
    }
}
