package uk.co.hcsoftware.commander.gui.ctable;

import org.jdesktop.swingx.JXTable;
import org.jdesktop.swingx.decorator.*;
import uk.co.hcsoftware.commander.AppProperties;
import uk.co.hcsoftware.commander.Main;
import uk.co.hcsoftware.commander.gui.CommanderFocus;
import uk.co.hcsoftware.commander.gui.util.ActionBuilder;
import uk.co.hcsoftware.commander.gui.util.GuiUtilities;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.prefs.Preferences;


/**
 * Title: CommanderTable
 * Copyright:    Copyright (c) 01-Mar-2006
 *
 * @author N Spurling
 */
public class CommanderTable extends JXTable{
    public static final Logger log = Logger.getLogger(CommanderTable.class.getName());
    public static final Dimension PREFERRED_SIZE = new Dimension(10,20);

    public CommanderTable(){
        this(null);
    }
    public CommanderTable(TableModel ctm) {
        super(ctm);
        log.fine("create commandertable model="+ctm);
        setHorizontalScrollEnabled(true);
        setColumnControlVisible(false);
        setShowGrid(false);
        setSortable(false); //disable default sorting behaviour: we will use something more sophisticated, with filters

        final JTableHeader th = getTableHeader();
        //setup sorting when click on headers
        th.setDefaultRenderer(new HeaderRenderer());
        final JPopupMenu popup = new JPopupMenu();
        popup.add(configureSorting);
        th.addMouseListener(new MouseAdapter(){
            private void mouse(MouseEvent e) {
                if(e.isPopupTrigger()){
                    popup.show(CommanderTable.this,e.getX(),e.getY());
                }

            }
            public void mousePressed(MouseEvent e) { mouse(e); }
            public void mouseReleased(MouseEvent e) { mouse(e); }

            public void mouseClicked(MouseEvent e) {
                if(SwingUtilities.isLeftMouseButton(e)){
                    int col = th.columnAtPoint(e.getPoint());
                    log.fine("mouseclicked col="+col);
                    headerClicked(col, (e.getModifiers() & AppProperties.getInstance().getCtrlMask())!= 0, e.isShiftDown());
                }
            }
        });
        //remove bindings for tab key (TAB should do nothing)
        GuiUtilities.removeBindings(this,KeyStroke.getKeyStroke(KeyEvent.VK_TAB,0));
        GuiUtilities.removeBindings(this,KeyStroke.getKeyStroke(KeyEvent.VK_TAB,KeyEvent.SHIFT_MASK));
        GuiUtilities.removeBindings(this,KeyStroke.getKeyStroke(KeyEvent.VK_F2,0));
        setupActions();

    }


    private void headerClicked(int viewColumn, boolean ctrlDown, boolean shiftDown) {
        int modelColumn = convertColumnIndexToModel(viewColumn);
        //if ctrldown && >1cols sorted
        if(ctrlDown || shiftDown) return; ///implement? not sure about this, shouldn't advanced func be only accessible from the right click..
        Sorter ss = filterManager.getSorter(modelColumn);//columnsToSorter.get(modelColumn);
        if(ss!=null){
            ss.toggle();
            getTableHeader().repaint();//repaint the header as sorting has changed
        }
        else{
            setSort(modelColumn);
        }
    }
    private FilterManager filterManager = new FilterManager();
    public static interface FilterFactory{
        public Filter getFilter();
    }
    /**
     * 3 kinds of filter:
     * pre-filters that are always applied
     * custom customSortersInOrderApplied (applied by user via gui)
     * post-filters that are always applied
     *
     * User configures customSortersInOrderApplied using SortTerms dialog
     */
    private class FilterManager{

        public Sorter getSorter(int column){
            return columnsToSorter.get(column);
        }
        public int getOrder(Sorter s){
            return customSortersInOrderApplied.indexOf(s);//.get(s);
        }
        private Map<Integer,Sorter> columnsToSorter = new HashMap<Integer,Sorter>();
        private List<Sorter> customSortersInOrderApplied=new ArrayList<Sorter>();
        List<ColumnSortDialog.SortTerm> sortTerms=Collections.emptyList();
        public void setColumnsToSorter(List<ColumnSortDialog.SortTerm> sortTerms){
            log.finer("setColumnsToSort terms="+sortTerms);
            this.sortTerms=sortTerms;
            columnsToSorter.clear();
            customSortersInOrderApplied.clear();
            for(ColumnSortDialog.SortTerm st:sortTerms){
                final int col = st.getColumnIndex();
                final boolean asc = st.isAscending();
                log.finer("sort by col="+col+" asc="+asc);
                ColumnSpec cs = getSpec(col);
                final ShuttleSorter ss  = new ShuttleSorter(col,asc);
                if(cs!=null && cs.hasComparator()){
                    ss.setComparator(cs.getComparator());
                }
                log.finer("no comparator defined for col="+col+": using default ShuttleSorter.comparator");
                customSortersInOrderApplied.add(ss);
                columnsToSorter.put(col,ss);
            }
        }
        //pre/post filters. 
        //however. 1. Don't think JXTable supports having multiple sorts per column (even if they use different comparators)
        //2. Can't re-apply the same Filter object, since applying it mutates it; after which it detects that it is being re-applied
        //(so might have to keep the data in a separate object and generate a Filter object each time we want to apply)
        private List<FilterFactory> preFilters=new ArrayList<FilterFactory>();
        public void addPreFilter(FilterFactory f){
            preFilters.add(f);
        }
        private List<FilterFactory> postFilters=new ArrayList<FilterFactory>();
        public void addPostFilter(FilterFactory f){
            postFilters.add(f);
        }
        public Filter[] getFilters() {
            Filter[]   filters = new Filter[customSortersInOrderApplied.size()+preFilters.size()+postFilters.size()];
            int filterIndex=0;
            for(FilterFactory ff:preFilters) filters[filterIndex++]=ff.getFilter();
            for(Sorter s:customSortersInOrderApplied) filters[filterIndex++]=s;
            for(FilterFactory ff:postFilters) filters[filterIndex++]=ff.getFilter();
            log.fine("return filters.length="+filters.length+" pre="+preFilters.size()+" cus="+customSortersInOrderApplied.size()+" post="+postFilters.size());
            return filters;
        }

        public boolean customSortIsMultiLevel() {
            return customSortersInOrderApplied.size()>1;
        }

        public List<ColumnSortDialog.SortTerm> getCustomSortTerms() {
            return sortTerms;
        }
    }

    private Action configureSorting = ActionBuilder.getInstance()
            .init(new AbstractAction("Advanced sorting..") {
                private ColumnSortDialog csd;
                public void actionPerformed(ActionEvent e) {
                    //we sort on the _model_ columns (not the visible columns) since the user might want to sort by a hidden column!
                    int colCount=getModel().getColumnCount();
                    List<String> colNames = new ArrayList<String>(colCount);

                    TableModel tm = getModel();
                    for(int i=0;i<colCount;i++){
                        if(i==0) colNames.add("Mark");
                        else colNames.add(tm.getColumnName(i));
                    }

                    Window w = SwingUtilities.getWindowAncestor(CommanderTable.this);
                    Frame f=null;
                    if(w instanceof Frame) f=(Frame) w;
                    if(csd==null){
                        csd = new ColumnSortDialog(f,colNames);
                        if(columnSpecs!=null){
                            List<SortDirection> sds = new ArrayList<SortDirection>(columnSpecs.size());
                            for(ColumnSpec cs:columnSpecs){
                                sds.add(cs.getDefaultSortDirection());
                            }
                            csd.setDefaultSortDirections(sds);
                        }
                    }
                    List<ColumnSortDialog.SortTerm> sortTerms = filterManager.getCustomSortTerms();
                    csd.setTerms(sortTerms);
                    csd.setLocationRelativeTo(CommanderTable.this);
                    List<ColumnSortDialog.SortTerm> terms = csd.showDialog();
                    if(terms==null) return;
                    log.fine("fetched terms="+terms);
                    filterManager.setColumnsToSorter(terms);
                    applySort();
                }
            }).tooltip("Sort by one or more columns..")
            .get();

    public void setSort(int modelColumn) {
        ColumnSpec cs = getSpec(modelColumn);
        final SortDirection dir = cs!=null?cs.getDefaultSortDirection():SortDirection.ascending;
        filterManager.setColumnsToSorter(Collections.singletonList(new ColumnSortDialog.SortTerm(dir,modelColumn)));
        applySort();
    }
    public void addPreFilter(FilterFactory f){
       filterManager.addPreFilter(f);
        applySort() ;
    }
    public void addPostFilter(FilterFactory f){
        filterManager.addPostFilter(f);
        applySort()  ;
    }
    private void applySort(){
        FilterPipeline pipeline =new FilterPipeline(filterManager.getFilters());
        getTableHeader().repaint();//repaint the header as sorting has changed
        setFilters(pipeline);
    }

    private void setupActions() {
        registerAction(firstRow, KeyStroke.getKeyStroke(KeyEvent.VK_HOME,0), "fr");
        registerAction(lastRow,KeyStroke.getKeyStroke(KeyEvent.VK_END,0),"lr");
    }


    public void registerAction(Action a, KeyStroke ks, String key){
        getInputMap(JComponent.WHEN_FOCUSED).put(ks,key);
        getActionMap().put(key,a);
    }
    public void registerAction(Action... actions){
        for (Action a:actions) {
            final KeyStroke ks = (KeyStroke)a.getValue(Action.ACCELERATOR_KEY);
            final String command = (String) a.getValue(Action.ACTION_COMMAND_KEY);
            if(ks==null || command==null) throw new IllegalArgumentException("action must have "+Action.ACCELERATOR_KEY+" and "+Action.ACTION_COMMAND_KEY
            + " values defined");
            registerAction(a,ks,command);
        }
    }

    private Action firstRow = new AbstractAction("home") {
        public void actionPerformed(ActionEvent e) {
            selectRow(0);
        }
    };
    Action lastRow = new AbstractAction("end") {
        public void actionPerformed(ActionEvent e) {
            final int ro = getRowCount()-1;
            if (ro>-1) {
                selectRow(ro);
            }
        }
    };


     /**
     *
     * @param scrollToVisibleIndex if -1 then ignore
     */
    public void selectRows(List<Integer> viewRows, int scrollToVisibleIndex) {
        ListSelectionModel lsm = getSelectionModel();
        lsm.clearSelection();
        for(int ro:viewRows){
            if (ro>-1) {
                lsm.addSelectionInterval(ro,ro);
            }
        }
        if(scrollToVisibleIndex>-1 && !viewRows.isEmpty()){
            scrollToVisible(scrollToVisibleIndex);
        }
    }
    public void selectRow(int viewRow) {
        if(getRowCount()>viewRow){
            log.finer("selectrow="+viewRow);
            getSelectionModel().setSelectionInterval(viewRow,viewRow);
            scrollToVisible(viewRow);
        }
        else log.info("ingoring selectRow as viewRow="+viewRow+" rocount="+getRowCount());
    }
    public void scrollToVisible(int row) {
        final Rectangle cellRect = getCellRect(row,0,true);
        log.finer("scrolltovisible cellRect="+cellRect);
        scrollRectToVisible(cellRect);
    }
    public boolean isSelectedFiles(boolean includeSelected, boolean includeMarkedFiles){
        if(includeSelected && getSelectedRowCount()>0){
            return true;
        }
        if(includeMarkedFiles){
            int count = getRowCount();
            for(int ro=0;ro<count;ro++){
                if(((Boolean)getValueAt(ro,0)).booleanValue()){
                    return true;
                }
            }
        }
       return false;
    }


    /**
     * set the owner of this component - sets and adjusts the highlighting coloration based on whether or not the owner has focus
     * @param cf
     * @param commanderFocusable
     */
    public void setCommanderFocusOwner(final CommanderFocus cf, final Component commanderFocusable){
        //add a highlighter to draw selected rows, need to use a highlighter so can avoid drawing selected rows if
        //we don't have the commanderfocus
        Highlighter[] hl = {new Highlighter(){
            protected Color computeBackground(Component renderer, org.jdesktop.swingx.decorator.ComponentAdapter adapter) {
                if(!cf.hasFocus(commanderFocusable)){
                    return CommanderTable.this.getBackground();
                }
                if(adapter.isSelected()) return CommanderTable.this.getSelectionBackground();
                return CommanderTable.this.getBackground();
            }
            protected Color computeForeground(Component renderer, org.jdesktop.swingx.decorator.ComponentAdapter adapter) {
                if(!cf.hasFocus(commanderFocusable)){
                    return CommanderTable.this.getForeground();
                }
                if(adapter.isSelected()) return CommanderTable.this.getSelectionForeground();
                return CommanderTable.this.getForeground();
            }
        }
        };
        CommanderTable.this.setHighlighters(new HighlighterPipeline(hl));
        //listen for the table getting the focus and notify the CommanderFocus manager
        addFocusListener(new FocusAdapter() {
            public void focusGained(FocusEvent e) {
                cf.giveFocus(commanderFocusable);
            }
        });
    }


    private class HeaderRenderer extends JLabel implements TableCellRenderer {
        public HeaderRenderer() {
            super.setHorizontalAlignment(JLabel.CENTER);
            setPreferredSize(PREFERRED_SIZE);
            setBorder(BorderFactory.createLineBorder(Color.black));
        }
        private String getHeaderText(String colName, int colIndex){
            StringBuilder sb = new StringBuilder();
            sb.append("<html><body><font=lucida>");
            Sorter s =filterManager.getSorter(colIndex);//columnsToSorter.get(colIndex);
            if (s!=null) {
//                    sb.append("<gtOne>");
                    final boolean gtOne = filterManager.customSortIsMultiLevel();//.size()>1;
                    if(gtOne){
                        sb.append("(");
                        sb.append(filterManager.getOrder(s)+1);//.get(colIndex)+1);
                    }

                    sb.append(s.isAscending()?"\u25b2":"\u25bc");//"\u25b4":"\u25be");//"\u25b2":"\u25bc");
//                    sb.append("</gtOne>");
                    if(gtOne) sb.append(")");
                    sb.append(" ");
            }
            sb.append(colName);
            sb.append("</body></html>");
            return sb.toString();
        }
        public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected, boolean hasFocus, int row, int column) {
            final String text = getHeaderText((String) value,column);
//            log.fine("text="+text);
            setText(text);
            return this;
        }
    }



    private class TableAction extends AbstractAction{
        private Action tableAction;
        public TableAction(String name, Action tableAction) {
            super(name);
            this.tableAction = tableAction;
        }
        public void actionPerformed(ActionEvent e) {
            requestFocusInWindow();
            e.setSource(CommanderTable.this);
            tableAction.actionPerformed(e);
        }
    }
     /**
     * find the relevant action in our inputmap and return an action
     * that 1) requests focus for the table 2) executes that action to the table
     * @param ks
      */
     public Action getForwardAction(int inputMapcondition, final KeyStroke ks) {
         try {
             ActionMap tableMap = getActionMap();
             InputMap table_im = getInputMap(inputMapcondition);
             String key = (String) table_im.get(ks);
             log.finer("tableky="+key);
             Action tableUp=tableMap.get(key);
             log.finer("tableup="+tableUp);
             return new TableAction(key,tableUp) ;
         } catch (Exception e) {
             log.log(Level.WARNING
                     ,"couldn't map key " +
                     "ks="+ks
                     ,e);
         }
         return null;
    }



    private ColumnSpec getSpec(int modelColumn){
        if(columnSpecs==null) return null;
        if(modelColumn<0 || modelColumn>=columnSpecs.size()){
            return null;
        }
        return columnSpecs.get(modelColumn);
    }
    private List<ColumnSpec> columnSpecs;
    private static final Preferences columnSizePrefs = Preferences.userNodeForPackage(Main.class);
    /**
     *
     * @param columnSpecs
     */
    public void configureColumns(final List<ColumnSpec> columnSpecs) {
        this.columnSpecs=columnSpecs;
        final TableColumnModel cm = getColumnModel();
        for (ColumnSpec cs:columnSpecs) {
            final TableColumn column = cm.getColumn(cs.getModelIndex());
            log.finer("configuring column="+cs);
            int w = columnSizePrefs.getInt(cs.getPrefKey(),cs.getDefaultWidth());
            if(w<2){
                log.warning("prefval for cs="+cs);
                w=cs.getDefaultWidth();// defaultWidth;
            }
            log.finer("setting colsize for cs="+cs+" to w="+w);
            column.setPreferredWidth(w);
            //init the sizes

            //attach any renderers
            final TableCellRenderer renderer = cs.getRenderer();
            if(renderer!=null){
                column.setCellRenderer(renderer);
            }
        }

        //listen for size changes


        cm.addColumnModelListener(new TableColumnModelListener() {
            boolean saveThreadRunning;
            final Thread saveThread = new Thread(){
                public void run() {

                    while (true) {
                        try {
                            if (saveTasksScheduled==0)
                                synchronized (this) {
                                    wait();
                                }
                            int scheduledTasks;
                            do{
                                scheduledTasks=saveTasksScheduled;
                                sleep(2000);
                            } while(scheduledTasks!=saveTasksScheduled);  //sleep until we have a clear 2seconds in which no further saves were scheulded
                            //now save the widths
                            saveTasksScheduled=0;
                            log.fine("saving columnwidths");
                            final TableColumnModel columnModel = getColumnModel();
                            for(ColumnSpec cs:columnSpecs){
                                final TableColumn column = columnModel.getColumn(cs.getModelIndex());
                                int width= column.getWidth();
                                columnSizePrefs.putInt(cs.getPrefKey(),width);
                            }

                        } catch (InterruptedException e) {
                            log.log(Level.WARNING, "exception", e);
                        }
                    }
                }
            };

            private volatile int saveTasksScheduled =0;
            public void columnAdded(TableColumnModelEvent e) { }
            public void columnRemoved(TableColumnModelEvent e) { }
            public void columnMoved(TableColumnModelEvent e) {  }
            public void columnMarginChanged(ChangeEvent e) {
                synchronized (saveThread) {
                    if(!saveThreadRunning){
                        saveThread.setDaemon(true);
                        saveThread.start();
                        saveThreadRunning=true;
                    }
                }
                //is there some way of ignoring 'intermediate' events while the user is dragging the column around?
                //SOLUTION: schedule the save action to 1s after the user moves the column. If any later tasks
                //have been scheduled in the period between the action and the task running, discard the tasks (as the final one supersedes)
                if (!columnSpecs.isEmpty()) {
                    saveTasksScheduled++;
                    synchronized (saveThread) {
                        saveThread.notify();
                    }
                }
            }
            public void columnSelectionChanged(ListSelectionEvent e) { }
        });
    }

    public void tableChanged(TableModelEvent e) {
        super.tableChanged(e);
        fireTableChanged(e);
    }

    private void fireTableChanged(TableModelEvent e) {
        if (tml!=null) {
            List<TableModelListener> tmlClone=new ArrayList<TableModelListener>(tml);
            for(TableModelListener tml:tmlClone){
                tml.tableChanged(e);
            }
        }
    }

    private List<TableModelListener> tml=new ArrayList<TableModelListener>();
    /**
     * receive notification of tablemodel events <em>after</em> they have been handled by the table (e.g. it has updated its internal state such as rowcount)
     * @param tml
     */
    public void addTableModelListener(TableModelListener tml){
        this.tml.add(tml);
    }
    public void removeTableModelListener(TableModelListener tml){
        this.tml.remove(tml);
    }

}
