package uk.co.hcsoftware.commander.gui.file;

import uk.co.hcsoftware.commander.gui.ProgressUpdater;

import javax.swing.*;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

public abstract class CopyMoveOperation {
    private static final Logger log = Logger.getLogger(CopyMoveOperation.class.getName());

    protected File destination;
    protected List<File> files;
    protected JComponent jc;
    protected List<CopyFailure> failures;
    private List<CopyDescriptor> schedule;
    private ProgressUpdater pu;


    public int countOverwrites() {
        List<CopyDescriptor> sched = getSchedule();
        int i=0;
        for(CopyDescriptor f:sched){
            if(f.getTo().exists()) i++;
        }
        log.fine("countoverwrites="+i);
        return i;
    }

    protected CopyMode getCopyMode(final File sourceFile, final File destFile, CopyMode copyMode){
        if(copyMode!=CopyMode.overwrite && destFile.exists()){  //overwrite is default
            log.fine("copymode !overwrite destfile="+destFile+" exists");
            if(copyMode==CopyMode.ask){
                final CopyMode[] holder = new CopyMode[1];
                Runnable runnable = new Runnable() {
                    public void run() {
                        final int i=JOptionPane.showConfirmDialog(jc,"overwrite "+destFile.getName()+"?","Copy",JOptionPane.YES_NO_CANCEL_OPTION);
                        switch(i){
                            case JOptionPane.YES_OPTION:
                                log.fine("overwrite? was an affirmative");
                                holder[0]=CopyMode.overwrite;
                                break;
                            case JOptionPane.NO_OPTION:
                                log.fine("overwrite? was a negatory, skipping directory="+sourceFile);
                                holder[0] =CopyMode.skip;
                                break;
                            default:
                                log.fine("ovewrite? was a cancel, abort method");
                                holder[0] = CopyMode.abort;//cancel
                        }
                    }
                };
                if (!SwingUtilities.isEventDispatchThread()) {
                    try {
                        SwingUtilities.invokeAndWait(runnable);
                    } catch (InterruptedException e) {
                        log.log(Level.WARNING,"interruped",e);
                    } catch (InvocationTargetException e) {
                        log.log(Level.WARNING,"invocationtarget",e);
                    }
                }
                else runnable.run();
                return holder[0];
            }
            else if(copyMode==CopyMode.skip){
                log.fine("copymode skip, skipping directory="+sourceFile);
                return CopyMode.skip;
            }
            else if(copyMode==CopyMode.overwrite_if_older && destFile.lastModified() >= sourceFile.lastModified()){
                log.fine("skipping file="+sourceFile+" as destFile modified after (or =) than sourcefile: " +
                        "destFile.lastModified()="+destFile.lastModified()+
                        " sourceFile.lastModified()="+sourceFile.lastModified());
                return CopyMode.skip;
            }
            else if(copyMode==CopyMode.abort){
                return CopyMode.abort;
            }
        }
        return CopyMode.overwrite;
    }

    public List<File> getFiles() {
        return files;
    }

    public File getDestination() {
        return destination;
    }

    public List<CopyFailure> getFailures() {
        return failures;
    }
    public void setProgressUpdater(ProgressUpdater pu){
        this.pu=pu;
    }
    protected void setStatusText(String s){
        if(pu!=null) pu.updateBlockMessage(s);
    }

    public static enum CopyMode{
        ask("ask"),skip("skip"),abort("abort"),overwrite("overwrite"),overwrite_if_older("overwrite if older");
        private String name;
        CopyMode(String name) {
            this.name = name;
        }
        public String getName() {
            return name;
        }

        public String toString() {
            return name;
        }

        public static CopyMoveOperation.CopyMode defaultCopyMode() {
            return overwrite_if_older;
        }
    }

    public static class CopyDescriptor{
        File from,to;

        public CopyDescriptor(File from, File to) {
            this.from = from;
            this.to = to;
        }

        public File getFrom() {
            return from;
        }

        public File getTo() {
            return to;
        }

        public String toString() {
            return "from="+from+" to="+to;
        }
    }

    public static class CopyFailure{
        private CopyDescriptor copyDescriptor;
        private IOException exception;
        private String s;

        public CopyFailure(CopyDescriptor copyDescriptor, IOException exception) {
            this.copyDescriptor = copyDescriptor;
            this.exception = exception;
        }

        public CopyFailure(CopyDescriptor copyDescriptor, String s) {
            this.copyDescriptor = copyDescriptor;
            this.s = s;
        }

        public CopyDescriptor getCopyDescriptor() {
            return copyDescriptor;
        }

        public String getReason(){
            if(exception!=null) return exception.getMessage();
            else if(s!=null) return s;
            else return null;
        }

        public String toString() {
            return "cd="+copyDescriptor+" exception="+exception;
        }
    }

    public CopyMode getCopyModeWithPrompt(String title) {
        final int overwrites = countOverwrites();//countOverwrites(files,directory);
        CopyOperation.CopyMode cm = CopyMode.defaultCopyMode();
        if (overwrites>0) {
            cm = promptForCopyMode(title);
        }
        return cm;
    }

    private CopyOperation.CopyMode promptForCopyMode(String title){
        final CopyOperation.CopyMode[] options = CopyOperation.CopyMode.values();
        int i=JOptionPane.showOptionDialog(jc
                ,"What do you want to do when a file already exists?"
                ,title
                ,JOptionPane.OK_CANCEL_OPTION
                ,JOptionPane.QUESTION_MESSAGE
                ,null,options,CopyOperation.CopyMode.defaultCopyMode());
        if(i>-1){
            return options[i];
        }
        else return null;
    }

    public List<CopyDescriptor> getSchedule(){
        if (schedule==null) {
            schedule=buildSchedule();
        }
        return schedule;
    }
    protected abstract List<CopyDescriptor> buildSchedule();
}
