package uk.co.hcsoftware.commander.gui.file;

import ns.gui.SwingWorker3;
import uk.co.hcsoftware.commander.gui.InputBlocker;
import uk.co.hcsoftware.commander.gui.util.ActionBuilder;
import uk.co.hcsoftware.commander.io.FileEventManager;
import uk.co.hcsoftware.commander.io.FileUtils;
import uk.co.hcsoftware.commander.io.DeleteProgress;
import uk.co.hcsoftware.commander.io.MoveProgress;
import uk.co.hcsoftware.commander.Main;

import javax.swing.*;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.prefs.Preferences;
import java.util.logging.Level;
import java.util.logging.Logger;


/**
 * Title: DeleteAction
 * Copyright:    Copyright (c) 01-Mar-2006
 *
 * @author N Spurling
 */
public class DeleteAction extends AbstractAction implements DeleteProgress {
    private static final Logger log = Logger.getLogger(DeleteAction.class.getName());
    private static final Preferences prefs = Preferences.userNodeForPackage(Main.class);

    private final Component host;
    private InputBlocker inputBlocker;
    private final FileSelector fs;
    private List<File> failures, deleted,changed,created;


    public DeleteAction(Component host, InputBlocker directoryPanel, FileSelector fs) {
        this.host = host;
        this.inputBlocker = directoryPanel;
        this.fs = fs;
        ActionBuilder.getInstance().init(this).keyStroke(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE,0)).command("delete");
    }

    public void actionPerformed(ActionEvent e) {
        java.util.List<File> selected = fs.getSelectedFiles(true,true,false);
        if(selected.isEmpty()) return;

        int fileCount = selected.size();
        String s = fileCount !=1?fileCount+" files":selected.get(0).getName();

        String softDeleteDir = prefs.get("softDeleteDir",FileUtils.defaultTrashDir().getAbsolutePath());
        // prompt (and get overwrite question)
        Object[] possibleValues = { "Soft Delete","Hard Delete", "Cancel" };
        int selectedValue = JOptionPane.showOptionDialog(host,
                "Delete "+s+"?"
                        +"\n   Soft means move files to:"
                        +"\n    " + softDeleteDir+
                         "\n   Hard means they will be lost forever!"
                ,"Delete"
                ,JOptionPane.CANCEL_OPTION,JOptionPane.QUESTION_MESSAGE, null,
                possibleValues, possibleValues[0]);
        log.finer("selval="+selectedValue);
        if(selectedValue<0 || selectedValue>1){//user cancelled or something
            return;
        }
        assert selectedValue==1 || selectedValue ==0;
        delete(selected, selectedValue==1);
    }

     private String getText(int toCopy) {
        if(toCopy==1) return "1 file";
        return (toCopy)+" files";
    }
    private String getMessage(){
        String s = "Deleted: "+getText(deleted.size())+"\n";
        if(!failures.isEmpty()) s+=  "Failed: "+getText(failures.size());
        return s ;

    }
    public void deleted(File f) {
        deleted.add(f);
        inputBlocker.updateBlockMessage(getMessage());
    }

    public void failedToDelete(File f) {
        failures.add(f);
        inputBlocker.updateBlockMessage(getMessage());
    }

    public void delete(final List<File> selected, final boolean hardNotSoft) {
        created = new ArrayList<File>();
        changed = new ArrayList<File>();
        deleted = new ArrayList<File>();
        failures = new ArrayList<File>();
        inputBlocker.blockInputWithCancel(getMessage());
        final String softDeleteDir = prefs.get("softDeleteDir",FileUtils.defaultTrashDir().getAbsolutePath());
        new SwingWorker3(){
            public Object construct() {
                try {
                    for(File f:selected){
                        if(hardNotSoft){
                          FileUtils.getInstance().deleteRecursive(f,DeleteAction.this);
                        }
                        else{// copy with structure (relative to root) to softDeleteDir
                            File newFile = createTrashPath(softDeleteDir, f);
                            final File destinationDir = newFile.getParentFile();
                            final boolean destExists = destinationDir.exists();
                            if (!destExists) {
                                final boolean b = destinationDir.mkdirs();
                                if (!b) {
                                    log.warning(destinationDir +" mkdirs failed!"); //this could be fine if destdir already exists
                                }
                            }
                            else log.finer("destinationdir exists already="+destinationDir);
                            boolean b = FileUtils.getInstance().moveRecursive(f, newFile);
                            if(b){
                                if(!destExists) created.add(newFile);
                                else changed.add(newFile);
                                deleted.add(f);
                            }
                            else{
                                failures.add(f);
                            }
                        }
                    }
                } catch (Exception e) {
                    return e;
                }
                return null;
            }

            private File createTrashPath(String softDeleteDir, File f) {
                String path = f.getAbsolutePath();
                //replace any funny characters
                path=path.replaceAll(":","");
                path=path.replaceAll("/{2,}","/");
                File newFile = new File(softDeleteDir,path);
                log.finer("trashpath for path="+path+" is="+newFile.getAbsolutePath());
                return newFile;
            }

            public void finished() {
                inputBlocker.unblockInput();
                if (log.isLoggable(Level.FINEST)) {
                    log.finest("successfully delted="+deleted);
                }
                log.finer("delsize="+deleted.size()+" csize="+created.size()+" chsize="+changed.size());
                FileEventManager.getInstance().fireEvent(deleted,created,changed);
                Object ret = get();
                if(ret!=null && ret instanceof Exception){
                    Exception e=(Exception) ret;
                    log.log(Level.WARNING,"error in delete",e);
                    JOptionPane.showMessageDialog(host,"Error during delete operation\n  "+e.getMessage(),
                            "Delete",
                            JOptionPane.ERROR_MESSAGE
                    );
                    return;
                }
                if(!failures.isEmpty()){
                    final String message="couldn't delete "+
                            (failures.size()>1
                            ?(failures.size()+" files")
                            :failures.get(0).getName());
                    JOptionPane.showMessageDialog(host,message,
                            "Delete",
                            JOptionPane.WARNING_MESSAGE
                    );
                }
                if(failures.size()>0){ //select failures
                    fs.selectFiles(failures);
                }
            }
        }.start();
    }

}
