package uk.co.hcsoftware.commander.gui.file;

import uk.co.hcsoftware.commander.gui.ctable.MarkableTableModel;
import uk.co.hcsoftware.commander.gui.util.WeakListener;
import uk.co.hcsoftware.commander.io.FileEventManager;

import javax.swing.*;
import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.ListIterator;
import java.util.logging.Logger;


/**
 * used by filesearch panel represents a flat list of files
 *
 * Title: FileTableModel
 * Copyright:    Copyright (c) 14-Mar-2006
 *
 * @author N Spurling
 */
public class FileListTableModel extends MarkableTableModel<File> implements FileEventManager.FileEventListener {
    private static final Logger log = Logger.getLogger(FileListTableModel.class.getName());

    public static final int MARK=0;
    public static final int NAME=1;
    public static final int PATH=2;
    public static final int SIZE=3;
    public static final int DATE=4;
    public static final int EXT=5;

    public FileListTableModel() {
        //listen for filesystem changes so we can update the view, use a weak listener so I can be gc'd
        WeakListener wl = new WeakListener(this, FileEventManager.FileEventListener.class){
            public Object onNullObserver() {
                FileEventManager.getInstance().removeListener((FileEventManager.FileEventListener) getProxy());
                return null;
            }
        };
        FileEventManager.getInstance().addListener((FileEventManager.FileEventListener)wl.getProxy());
    }

    public void setFiles(List<File> files) {
        setDir(null,files);
    }

    public void filesChanged(FilesChangedEvent fe) {
        List<File> deleted = fe.getDeleted();
        //nested search watch out if this gets too slow
        int min=-1,max=-1;
        final List<Node> nodes = new ArrayList<Node>(children);
        int i=nodes.size()-1;
        //go backwards so i doesn't get out of sync (since we are mutating the list)
        for (ListIterator<Node> iterator = nodes.listIterator(nodes.size()); iterator.hasPrevious();i--) {
            Node r = iterator.previous();
            File f=r.getElement();
            if(deleted.contains(f)){
                iterator.remove();
                if(min==-1 || i<min) min=i;
                if(max==-1 || i>max) max=i;
            }
        }
        nodesRemoved(min, max, nodes);
    }
    public void directoryChanged(File directory) {
        //nested search watch out if this gets too slow
        int min=-1,max=-1;
        final List<Node> nodes = new ArrayList<Node>(children);
        int i=nodes.size()-1;
        //go backwards so i doesn't get out of sync (since we are mutating the list)
        for (ListIterator<Node> iterator = nodes.listIterator(nodes.size()); iterator.hasPrevious();i--) {
            Node r = iterator.previous();
            File f=r.getElement();
            if(directory.equals(f.getParentFile()) && !f.exists()){
                iterator.remove();
                if(min==-1 || i<min) min=i;
                if(max==-1 || i>max) max=i;
            }
        }
        nodesRemoved(min, max, nodes);

    }

    private void nodesRemoved(int min, int max, final List<Node> nodesNow) {
        if(min!=-1 && max!=-1){
            final int min1 = min;
            final int max1 = max;
            final Runnable runnable = new Runnable() {
                public void run() {
                    children=nodesNow;
                    fireTableRowsDeleted(min1,max1);
                }
            };
            if(SwingUtilities.isEventDispatchThread()) runnable.run();
            else SwingUtilities.invokeLater(runnable);
        }
    }


    public int getColumnCount() {return  6;}

    public String getColumnName(int index) {
        switch(index){
            case MARK: return "";
            case NAME: return "Name";
            case PATH: return "Path";
            case SIZE: return "Size";
            case DATE: return "Date";
            default: return "Ext";
        }
    }
    public Object getValueAt(int rowIndex, int attribute) {
        Node n = children.get(rowIndex);
        File f=n.getElement();
        switch(attribute){
            case MARK: return n.isMarked();
            case NAME: return f.getName();
            case PATH: final File parentFile = f.getParentFile();
                return parentFile==null?"":parentFile.getPath();
            case SIZE: return f.isDirectory()?-1:f.length();// hack to signal to filters that this is a dir
            case DATE: return f.lastModified();
            default: if(f.isDirectory()) return "";
                String nam = f.getName();
                int i = nam.lastIndexOf('.');
                if(i==-1 || i==nam.length()-1) return null;
                return nam.substring(i+1);
        }
    }


}
