package uk.co.hcsoftware.commander.gui.file;

import ns.gui.SwingWorker3;
import uk.co.hcsoftware.commander.gui.ctable.MarkableTableModel;
import uk.co.hcsoftware.commander.gui.util.WeakListener;
import uk.co.hcsoftware.commander.gui.ExceptionHandler;
import uk.co.hcsoftware.commander.io.FileEventManager;
import uk.co.hcsoftware.commander.io.DirectoryWatcher;

import javax.swing.*;
import java.io.File;
import java.util.*;
import java.util.logging.Logger;
import java.util.logging.Level;


/**
 * Title: FileTableModel
 * Copyright:    Copyright (c) 14-Mar-2006
 *
 * @author N Spurling
 */
public class FileTableModel extends MarkableTableModel<Vfile> implements FileEventManager.FileEventListener {
    private static final Logger log = Logger.getLogger(FileTableModel.class.getName());

    public static final int MARK=0;
    public static final int NAME=1;
    public static final int SIZE=2;
    public static final int DATE=3;
    public static final int EXT=4;

    public FileTableModel() {
        //listen for filesystem changes so we can update the view, use a weak listener so I can be gc'd
        WeakListener wl = new WeakListener(this, FileEventManager.FileEventListener.class){
            public Object onNullObserver() {
                FileEventManager.getInstance().removeListener((FileEventManager.FileEventListener) getProxy());
                return null;
            }
        };
        FileEventManager.getInstance().addListener((FileEventManager.FileEventListener)wl.getProxy());
    }
    public static final class CannotAccessDirectoryException extends Exception{
        private File dir;

        public CannotAccessDirectoryException(File dir) {
            this.dir = dir;
        }

        public File getDir() {
            return dir;
        }
    }
    /**
     *  WARNING: asynchronous (returns immediately, changes happen later)
     * @param dir
     * @param r runs in EDT when dir refreshed, can be null
     */
    public void setDirAsync(final Vfile dir, final Runnable r, final ExceptionHandler eh) {
//         * DirPanel doesn't look that good because most of the work on a slow drive is actually in the EDT, after setVisible(),
//         * when the table is querying all the files to get their attributes. If each of these methcalls takes an appreciable amount
//         * of time, then the GUI will freeze no matter what.
//         SO: we pre-query some of the file's attributes
//        problem with this is that the 1st 20 are not necessarily the 1st 20 visibles, because we do our own sorting/filtering in dirpanel
//        setDir(..) -> fireTableDataChanged(..) -> filter.reset(..) -> repaint(..)? all of which has to be on EDT
//        SO: we cheat a bit and cache the 1st 20 that are directories, since we know the sort always puts dirs first.


        log.finer("set setDirAsync="+dir);
        new SwingWorker3(){
            public Object construct() {
                try {
                    if(dir!=null){
                        final Vfile[] vfiles = dir.listFiles();
                        if(vfiles==null) throw new CannotAccessDirectoryException(dir.getFile());
                        log.finer("pre-caching dir attributes");
                        for(int i=0,j=0;j<20&&i<vfiles.length;i++){
                            if (vfiles[i].isDirectory()) {
                                vfiles[i].cacheProperties();
                                log.finer("cached dir="+vfiles[i].getName());
                                j++;
                            }
                        }
                        log.finer("done caching");
                        return Arrays.asList(vfiles);
                    }
                    else{
                        return Collections.emptyList();
                    }
                } catch (Exception e) {
                    log.log(Level.WARNING,"error",e);
                    return e;
                }
            }
            public void finished() {
                final Object o = get();
                if(o instanceof Exception){
                   eh.handleException ((Exception)o);
                    return;
                }
                setDir(dir,(List<Vfile>) o);
                if(dir!=null && dir.exists()){
                    final File fileDir = dir.getFile();
                    directoryWatcher = new DirectoryWatcher(fileDir) {
                        private List<File> getFileList(List<String> names){
                            List<File> l = new ArrayList<File>();
                            for(String n:names){
                                l.add(new File(fileDir,n));
                            }
                            return l;
                        }
                        final List<File> changed = Collections.emptyList();
                        public void changed(final List<String> added, final List<String> deleted) {
                            log.fine("firing fileschanged del.size()="+deleted.size()+" ad.size()="+added.size());
                            if(!SwingUtilities.isEventDispatchThread()){
                                SwingUtilities.invokeLater(new Runnable() {
                                    public void run() {
                                        changedEdt(added,deleted);
                                    }
                                });
                            }
                            else changedEdt(added,deleted);
                        }
                        private void changedEdt(final List<String> added, final List<String> deleted) {

                            filesChanged(new FilesChangedEvent(getFileList(deleted),getFileList(added),changed));
                        }
                    };
                }
                log.finer("finished setting dir on jxtable");
                if(r!=null){
                    r.run();
                }
            }
        }.start();
    }
    private class NodeSet{
        //assume all are sibling files
        private final HashMap<String,Node> nodes;
        public NodeSet(List<Node> nodes) {
            this.nodes=new HashMap<String, Node>(nodes.size());
            for(Node n:nodes) this.nodes.put(n.getElement().getName(),n);
//            this.nodes = nodes;
        }

        public Collection<Node> getNodes() {
            return nodes.values();
        }

        public int size() {
            return nodes.size();
        }

        /**
         *
         * @param deleted
         * @return count of nodes actually removed
         */
        public int removeAll(List<File> deleted) {
            int removed=0;
            for(File f:deleted){
                if(nodes.remove(f.getName())!=null){
                    removed++;
                }
//                for (Iterator<Node> iterator = nodes.iterator(); iterator.hasNext();) {
//                    if(iterator.next().getElement().sameFile(f)){
//                        iterator.remove();
//                        break;
//                    }
//                }
            }
            return removed;
        }

        /**
         *
         * @param added
         * @return count of nodes added
         */
        public int addAll(List<File> added) {
            int addedI=0;
            for(File f:added){
                if(nodes.put(f.getName(),new Node(Vfile.getVfile(f)))==null){
                    addedI++;
                }
            }
            return addedI;
        }

        /**
         *
         * @param changed
         * @return count of nodes replaced (excludes those that were added)
         */
        public int replaceAll(List<File> changed) {
            int added=addAll(changed);
            return changed.size()-added;
//            for(File f:changed){
//                 for (ListIterator<Node> iterator = nodes.listIterator(); iterator.hasNext();) {
//                     final Node node = iterator.next();
//                     if(node.getElement().sameFile(f)){
//                         node.getElement().resetCachedProperties();
//                         break;
//                     }
//                 }
//             }
        }
    }
    private DirectoryWatcher directoryWatcher;
    public void checkForChanges(){
        boolean nl=directoryWatcher==null;
        log.fine("checkForChanges directoryWatcher=null="+nl);
        if(!nl){
            directoryWatcher.checkForChanges();
        }
    }
    /**
     * if called in EDT, returns after model has been updated.. else updates model asynchronously
     * @param fe
     */
    public void filesChanged(FilesChangedEvent fe) {
        log.finer("filesChanged fe="+fe);
        if(dir==null){
            log.fine("ignoring event as directory==null");
            return;
        }
        File directoire = dir.getFile();
        final List<File> del = fe.getDeleted(directoire);
        final List<File> added = fe.getAdded(directoire);
        final List<File> changed = fe.getChanged(directoire);
        int changeCount = del.size()+added.size()+changed.size();
        log.fine("dir="+directoire+" changeCount="+changeCount);
        if(changeCount>0 && children!=null){
            final NodeSet nodeSet = new NodeSet(children);
            // now apply changes
            final int nodesRemoved=nodeSet.removeAll(del);
            final int nodesAdded= nodeSet.addAll(added);

            //replace the changed ones as the currs might have stale caches
            final int nodesChanged=nodeSet.replaceAll(changed);
            log.fine("children prevlen="+children.size()+" afterlen="+nodeSet.size());

//            forward changes to model listeners
            //don't bother trying to translate into fireUpdated(..) fireDeleted(..) etc. as those methods assume changes
            //are contiguous. It shouldn't be too much of a performance it to requery everything since the data is already
            //cached in memory (we edit the cache above)
            final Runnable runnable = new Runnable() {
                public void run() {
                    //this is here for complicated reasons: basically, we may get a dirupdated event saying this dir has had a file added
                    //while we are in the process of creating a file (e.g. a tar) so the table updates and sets the size to the size at the time it updates
                    //later the op finishes and fileeventmanager fires an event to say the file has been added
                    //when we update the table we notice the 'added' file is already there.
                    //We don't want to fire another tableDataChanged as it is inefficent
                    //so we just reset the cached props of all the added/changed nodes and tell the table taht those nodes have changed
                    if(nodesRemoved+nodesAdded+nodesChanged==0){
                        log.fine("not modifying children, as tablemodel was already up-to-date: contained all the nodes that were added/deleted/changed");
//                        return;
                        //go thru children and if any are in added/changed reset them
                        int min=-1,max=-1;
                        int i=0;
                        for(Node child:children){
                            final Vfile el = child.getElement();
                            if(added.contains(el.getFile()) || changed.contains(el.getFile())){
                                el.resetCachedProperties();
                                if(min==-1 || i<min) min=i;
                                if(max==-1 || i>max) max=i;
                            }
                            i++;
                        }
                        if(min!=-1 && max!=-1){
                            log.fine("children min="+min+" max="+max+" were changed: firing update");
                            fireTableRowsUpdated(min,max);
                        }
                    }
                    else{
                        children = new ArrayList<Node>(nodeSet.getNodes());
                        log.finer("set children len="+children.size());
                        fireTableDataChanged();
                        log.finer("completed fireTableDataChanged");
                    }
                }
            };
            if(SwingUtilities.isEventDispatchThread()){
                runnable.run();
            }
            else SwingUtilities.invokeLater(runnable);
        }
    }
    public void directoryChanged(File directory) {
        log.finer("received dirchanged dir="+directory);
         if(dir!=null && dir.sameFile(directory)){
             log.fine("refreshing view as directory has changed");
             setDirAsync(Vfile.getVfile(directory),null,new ExceptionHandler() {
                 public void handleException(Exception e) {
                     log.log(Level.WARNING,"exception setting dir async after dirchanged event",e);
                 }
             });
         }
    }



    public int getColumnCount() {return  5;}

    public String getColumnName(int index) {
        switch(index){
            case MARK: return "";
            case NAME: return "Name";
            case SIZE: return "Size";
            case DATE: return "Date";
            default: return "Ext";
        }
    }
    public Object getValueAt(int rowIndex, int attribute) {
        Node n = children.get(rowIndex);
        Vfile vfile=n.getElement();
        switch(attribute){
            case MARK: return n.isMarked();
            case NAME: return vfile.getName();
            case SIZE: return vfile.isDirectory()?-1:vfile.length();// hack to signal to filters that this is a dir
            case DATE: return vfile.lastModified();
            default: return vfile.getExtension();
        }
    }


}
