package uk.co.hcsoftware.commander.gui.file;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.nio.channels.FileChannel;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * virtual file, for now just wraps java.io.File and adds caching + more sane methods
 *
 * may be the basis of a VFS..  though probably need to change teh methods somewhat
 *
 *
 *  browsing around some NFS shares, the lion's share of time is spent in isDirectory (surprisingly)
 * but also in length() lastModified()
 * So we cache these properties
 *
 */
public class Vfile {
    public static final Logger log = Logger.getLogger(Vfile.class.getName());
    private final File file;

    //cached properties
    private long lm;
    private boolean lminit;
    private long len;
    private boolean leninit;
    private boolean isDir,dirInited;

    public void resetCachedProperties(){
        lminit=false; leninit=false; dirInited=false;
    }
    public void cacheProperties() {
        length();
        isDirectory();
        lastModified();
    }
    public String toString() {
        return file.toString()+" lmcached="+lminit+" lencached="+leninit+" isdircached="+dirInited;
    }
    public Vfile(String pathname) {
        file = new File(pathname);
    }
    public Vfile(URI uri) {
        file = new File(uri);
    }
    public Vfile(URL url) {
        file = new File(url.getFile());
    }

    public Vfile(Vfile f){
       this(f.file);
    }
    public static List<Vfile> getVfiles(List<File> l){
        List<Vfile> ll = new ArrayList<Vfile>(l.size());
        for(File f:l) ll.add(new uk.co.hcsoftware.commander.gui.file.Vfile(f));
        return ll;
    }


    public Vfile(File f){
        this.file=f;
        if(file==null) log.log(Level.WARNING,"f=null",new Throwable());
    }
    public Vfile(Vfile parent, String child) {
        file = new File(parent.file, child);
    }


    public Vfile getParentFile() {
        final File parentFile = file.getParentFile();
        return parentFile==null?null:new Vfile(parentFile);
    }

    public Vfile[] listFiles() {
        final File[] ss = file.listFiles();
        if (ss == null) return null;
        ArrayList v = new ArrayList(ss.length);
        for (int i = 0 ; i < ss.length ; i++) {
            Vfile f =  new Vfile(ss[i]);
            v.add(f);
        }
        return (Vfile[])(v.toArray(new Vfile[v.size()]));
    }

    public boolean mkdir() {
        return file.mkdir();
    }

    public boolean mkdirs() {
        return file.mkdirs();
    }

    public boolean renameTo(Vfile dest) {
        return file.renameTo(dest.file);
    }
    public void copyTo(final Vfile destFile) throws IOException {
        log.fine("copyfile src="+toString()+" dest="+destFile);
        FileChannel srcChannel = new FileInputStream(file).getChannel();

        try {
            FileChannel dstChannel = new FileOutputStream(destFile.file).getChannel();
            try {
                dstChannel.transferFrom(srcChannel, 0, srcChannel.size());
            } finally {
               dstChannel.close();
            }
        } finally {
            srcChannel.close();
        }
    }
    /**
     *
     * @param f
     * @return true if this vfile refers to the same file as f (according to File.equals) if either f or the file this Vfile refers to are null, return false
     */
    public boolean sameFile(File f) {
        if ((f != null && file!=null)) {
            return file.equals(f);
        }
        return false;
    }
    public boolean equals(Object obj) {
        if ((obj != null) && (obj instanceof Vfile)) {
            return file.equals(((Vfile)obj).file);
        }
        return false;
    }






    //query properties
    public URL toURL() throws MalformedURLException {
        return file.toURL();
    }
    public String getName() {
      //  long l1 = System.currentTimeMillis();
        final String name = file.getName();
       // getname+=(System.currentTimeMillis()-l1);
        return name;
    }
    public int hashCode() {
        return file.hashCode();
    }



    public String getAbsolutePath() {
        return file.getAbsolutePath();
    }
    public boolean exists() {
        return file.exists();
    }

    public boolean isDirectory() {
        if(!dirInited){
            isDir=file.isDirectory();
            dirInited=true;
        }
        return isDir;
    }


    public long lastModified() {
        if(!lminit){
            lm=file.lastModified();
            lminit=true;
        }
        return lm;
//        long l = System.currentTimeMillis();
//        final long lm = file.lastModified();
//        getlastmodified+=(System.currentTimeMillis()-l);
//        return lm;
    }


    public long length() {
         if(!leninit){
            len=file.length();
            leninit=true;
        }
        return len;
//        long l1 = System.currentTimeMillis();
//        final long l = file.length();
//        getlength+=(System.currentTimeMillis()-l1);
//        return l;
    }

    public File getFile() {
        return file;
    }

    public static List<File> getFiles(List<Vfile> d) {
        List<File> l = new ArrayList<File>(d.size());
        for(Vfile v:d) l.add(v.getFile());
        return l;
    }

    public String getExtension() {
        if(isDirectory()) return "";
        String n = getName();
        int i = n.lastIndexOf('.');
        if(i==-1 || i==n.length()-1) return null;
        return n.substring(i+1);
    }

    public boolean isFile() {
        return file.isFile();
    }

    public static Vfile getVfile(File file) {
        return new Vfile(file);
    }


    public static File[] getFilesArray(List<Vfile> d) {
        File[] l = new File[d.size()];
        int i=0;
        for(Vfile v:d) l[i++]=v.getFile();
        return l;
    }
}
