package uk.co.hcsoftware.commander.gui.usercommands;

import org.jdom.Element;
import uk.co.hcsoftware.commander.gui.CommanduleContext;

import java.io.File;
import java.util.*;
import java.util.logging.Logger;


/**
 * Title: UserCommand
 * Copyright:    Copyright (c) 20-Jul-2006
 *
 * @author N Spurling
 */
public class UserCommand implements Cloneable{
    private static final Logger log = Logger.getLogger(UserCommand.class.getName());
    private String id;
    private String name="<untitled>";
//    private String commandText;
    private String command;
    private List<String> args=new ArrayList<String>(); //cannot be null
    private boolean showOutputWindow=true;
    private boolean useFilesVariable=true, surroundWithQuotes=false,escapeSpaces=true,spaceDelimited=true;
    private boolean refreshDirectoryAfter;

    public UserCommand(String name) {
        this(generateId(),name);
    }

    private UserCommand(String id, String name) {
        this.id = id;
        this.name = name;
    }

    public String toString() {
        return "name="+name+" id="+id;
    }

    private static String generateId() {
        final Random r = new Random();
        return nextHex(r)+nextHex(r);
    }
    private static String nextHex(Random r){
        return Integer.toHexString(r.nextInt(Integer.MAX_VALUE));
    }

    public String getId() {
        return id;
    }

    protected Object clone() {
        try {
            final UserCommand clone = (UserCommand) super.clone();
            clone.args=new ArrayList<String>(args);
            return clone;
        } catch (CloneNotSupportedException e) {
            throw new RuntimeException("CloneNotSupported");
        }
    }
    /**
     *
     * @param sameId
     * @param name if null then copy existing name
     * @return
     */
    public UserCommand getCopy(boolean sameId, String name){
        UserCommand uc = (UserCommand) clone();
        if(!sameId) uc.id=generateId();
        if(name!=null) uc.name=name;
        return uc;
    }

    public void setCommand(String command) {
        this.command = command;
    }

    public void setArgs(List<String> args) {
        this.args = args;
    }

    public String getCommand() {
        return command==null?"":command;
    }

    public List<String> getArgs() {
        return args;
    }

    public void setName(String name) {
        this.name = name;
    }

    public boolean isRefreshDirectoryAfter() {
        return refreshDirectoryAfter;
    }

    public void setRefreshDirectoryAfter(boolean refreshDirectoryAfter) {
        this.refreshDirectoryAfter = refreshDirectoryAfter;
    }
//    public void setCommandText(String commandText) {
//        this.commandText = commandText;
//    }

    public void setShowOutputWindow(boolean showOutputWindow) {
        this.showOutputWindow = showOutputWindow;
    }

    public void setUseFilesVariable(boolean useFilesVariable) {
        this.useFilesVariable = useFilesVariable;
    }

    public void setSurroundWithQuotes(boolean surroundWithQuotes) {
        this.surroundWithQuotes = surroundWithQuotes;
    }

    public void setEscapeSpaces(boolean escapeSpaces) {
        this.escapeSpaces = escapeSpaces;
    }

    public void setSpaceDelimited(boolean spaceDelimited) {
        this.spaceDelimited = spaceDelimited;
    }

    public boolean isUseFilesVariable() {
        return useFilesVariable;
    }

    public String getName() {
        return name==null?"":name;
    }


    public boolean isShowOutputWindow() {
        return showOutputWindow;
    }

    public boolean isSurroundWithQuotes() {
        return surroundWithQuotes;
    }

    public boolean isEscapeSpaces() {
        return escapeSpaces;
    }

    public boolean isSpaceDelimited() {
        return spaceDelimited;
    }



    public ProcessBuilder createProcessBuilder(CommanduleContext sc) {
        ProcessBuilder pb = new ProcessBuilder();
        final File dir;
        if ((dir=sc.getDirectory())!=null) {
            pb.directory(dir);
        }
        else log.fine("not setting cwd as dir is null");

        if(useFilesVariable){  //set the FILES variable
            final StringBuilder sb = new StringBuilder();
            final List<File> selectedFiles = sc.getSelectedFiles();
            for(File f:selectedFiles){   // configure quoting
                String name1 = f.getName();
                if(surroundWithQuotes) sb.append("\"");
                if(escapeSpaces && name1.contains(" ")) name1=name1.replaceAll(" ","\\\\ ");
                sb.append(name1);
                if(surroundWithQuotes) sb.append("\"");
                if(spaceDelimited) sb.append(' ');
                else sb.append('\n'); //newline delimited
            }
            if(sb.length()>0){
                sb.setLength(sb.length()-1);
            }
            log.fine("export FILES variable="+sb.toString()+" (selectedfilecount="+selectedFiles.size()+")");
            pb.environment().put("FILES",sb.toString());//config variable name
        }

        //looking at processbuilder, the zeroth arg is the prog name,
        //everything else is a parameter and is quoted if not already quoted
        if(command==null || command.trim().length()==0){
            throw new IllegalStateException("Invalid command (cannot be blank)");
        }

        List<String> cmdList = new ArrayList<String>(args.size()+1);//parseCommandline(commandText);
        cmdList.add(command);
        cmdList.addAll(args);
        log.fine("creating processbuilder cmdList="+cmdList);
        pb.command(cmdList);
        return pb;
    }

    public boolean equals(Object obj) {
        if(!(obj instanceof UserCommand)) return false;
        UserCommand other =(UserCommand) obj;
        return this.id.equals(other.id);
    }

    public int hashCode() {
        return id.hashCode();
    }

    public Element writeToElement() {
        Element element = new Element("usercommand");
        element.setAttribute("id",id);
        element.setAttribute("name",name);
        Element cmd = new Element("command");
        element.addContent(cmd);
        cmd.setAttribute("commandName",command==null?"":command);
        int index=0;
        for(String arg:args){
            if (!isNull(arg)) {
                final Element argElement = new Element("argument");
                argElement.setAttribute("index",String.valueOf(index));
                argElement.setAttribute("value",arg);
                cmd.addContent(argElement);
                index++;
            }
        }
        element.setAttribute("showOutputWindow",String.valueOf(showOutputWindow));
        element.setAttribute("useFilesVariable",String.valueOf(useFilesVariable));
        element.setAttribute("surroundWithQuotes",String.valueOf(surroundWithQuotes));
        element.setAttribute("escapeSpaces",String.valueOf(escapeSpaces));
        element.setAttribute("spaceDelimited",String.valueOf(spaceDelimited));
        element.setAttribute("reloadDirAfter",String.valueOf(refreshDirectoryAfter));
        return element;
    }

    private boolean isNull(String arg) {
        return arg==null || arg.trim().length()==0;
    }

    public static UserCommand readFromElement(Element element) {
        UserCommand uc = new UserCommand(element.getAttributeValue("id"),element.getAttributeValue("name"));
        Element cmd = element.getChild("command");
        if (cmd!=null) {
            uc.command=cmd.getAttributeValue("commandName");
            List args = cmd.getChildren("argument");
            uc.args=new ArrayList<String>(args.size());
            for (Iterator iterator = args.iterator(); iterator.hasNext();) {
                Element argElement = (Element) iterator.next();
                String val = argElement.getAttributeValue("value");
                uc.args.add(val);
            }
        }

        uc.showOutputWindow=getBoolean(element, "showOutputWindow", uc.showOutputWindow);
        uc.useFilesVariable=getBoolean(element, "useFilesVariable", uc.useFilesVariable);
        uc.surroundWithQuotes=getBoolean(element, "surroundWithQuotes", uc.surroundWithQuotes);
        uc.escapeSpaces=getBoolean(element, "escapeSpaces", uc.escapeSpaces);
        uc.spaceDelimited=getBoolean(element, "spaceDelimited", uc.spaceDelimited);
        uc.refreshDirectoryAfter=getBoolean(element,"reloadDirAfter",uc.refreshDirectoryAfter);
        return uc;
    }

    private static boolean getBoolean(Element element, String attName, boolean defaultVal) {
        String s = element.getAttributeValue(attName);
        if(s==null || s.trim().length()==0){
            return defaultVal;
        }
        return Boolean.valueOf(s);  //returns false if illegal string (judgin by src)
    }

    public String getFullCommandString() {
        List<String> cmdList = new ArrayList<String>(args.size()+1);//parseCommandline(commandText);
        cmdList.add(command);
        cmdList.addAll(args);
        StringBuilder sb = new StringBuilder();
        for(String s:cmdList){
            sb.append(s);
            sb.append(' ');
        }
        if(sb.length()>0){
            sb.setLength(sb.length()-1);
        }
        return sb.toString();
    }

    
}
