package uk.co.hcsoftware.commander.gui.util;

import javax.swing.*;
import java.util.logging.Logger;


/**
 * Title: ActionBuilder
 * Copyright:    Copyright (c) 09-Mar-2006
 *
 *
 * Version of the 'Builder' pattern
 *
 * basically you write code that looks something like :
 *
 * <pre>
   ActionBuilder ab = new ActionBuilder();
   Action refeshAction = ab.init(new AbstractAction("") {
        public void actionPerformed(ActionEvent e) {
            refresh();
        }
    })
            .icon(ImageProvider.getInstance().getIcon("icons/16/reload.png"))
            .tooltip("Reload current directory from filesystem")
            .keyStroke(KeyStroke.getKeyStroke(KeyEvent.VK_R,AppProperties.getInstance().getCtrlMask()))
            .get();</pre>


 * <p>this means you can initialise all the various properties of Action in the declaration,
 * without having to have 1 constructor (or factory method)
 * per combination that you want to use.</p>
 *
 * <h3>So it gives the benefits:</h3>
 * <ul><li>Initialise in declaration</li><li>Avoid lots of boilerplate factory methods</li><ul>
 * <h3>In exchange for:</h3>
 * <ul><li>hideous-looking, non-obvious code (unless you are familiar with the pattern)</li></ul>
 *
 * Time will tell if it's worth it.
 *
 *
 * @author N Spurling
 */
public class ActionBuilder {
    public static final Logger log = Logger.getLogger(ActionBuilder.class.getName());
//    private static ActionBuilder ab=new ActionBuilder();
    private static ThreadLocal<ActionBuilder> tl = new ThreadLocal<ActionBuilder>(){
        protected ActionBuilder initialValue() {
            return new ActionBuilder();
        }
    }; //one actionbuilder per thread

    public static ActionBuilder getInstance(){
        return tl.get();
    }
    private Action a;

    public ActionBuilder init(Action a){
        this.a=a;
        return this;
    }
    public ActionBuilder command(String command){
        a.putValue(Action.ACTION_COMMAND_KEY,command);
        return this;
    }
    public ActionBuilder mnemonicKey(int keyEvent) {
        a.putValue(Action.MNEMONIC_KEY,keyEvent);
        return this;
    }

    public ActionBuilder keyStroke(KeyStroke ks){
        a.putValue(Action.ACCELERATOR_KEY,ks);
        return this;
    }
    public ActionBuilder name(String actionName){
        a.putValue(Action.NAME,actionName);
        return this;
    }
    public ActionBuilder tooltip(String tooltip){
        a.putValue(Action.SHORT_DESCRIPTION,tooltip);
        return this;
    }
    public ActionBuilder icon(Icon ks){
        a.putValue(Action.SMALL_ICON,ks);
        return this;
    }

    public Action get(){
//        if(a.getValue(Action.ACCELERATOR_KEY)!=null && a.getValue(Action.ACTION_COMMAND_KEY)==null){
//            throw new IllegalStateException("This action has an accelerator defined but no action command,..")
//        }
        return a;
    }

}
