/**
 * Title: ${NAME}
 * Copyright:    Copyright (c) 08-Mar-2006
 * @author N Spurling
 */
package uk.co.hcsoftware.commander.gui.util;

import uk.co.hcsoftware.commander.gui.Commandule;
import uk.co.hcsoftware.commander.io.ImageProvider;
import uk.co.hcsoftware.commander.AppProperties;

import javax.swing.*;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.WeakHashMap;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * convenience class for generating all the typical commander actions (which each commandule needs a copy of)
 *
 * Uses reflection, to save typing, and creating multiple anonymous classes
 *
 * Stores commandules in a WeakHashMap to allow them to be gc'd
 *
 */
public class ActionDispatcher {
     public static final Logger log = Logger.getLogger(ActionDispatcher.class.getName());
    private static ActionDispatcher ourInstance = new ActionDispatcher();

    public static ActionDispatcher getInstance() {
        return ourInstance;
    }

    private ActionDispatcher() {
    }


    private Map<CommanduleSource,Commandule> actionsToCommandules = new WeakHashMap<CommanduleSource,Commandule>();
    public interface CommanduleSource{
        /**
         *
         * @return may be null, in which case nothing happens
         */
        Commandule getCommandule();
    }
    public List<Action> createCommanderActions(CommanduleSource cs){
        List<Action> actions = new ArrayList<Action>();
        ActionBuilder ab = ActionBuilder.getInstance();
        actions.add(ab.init(getAction("usercommand", cs))
                .keyStroke(KeyStroke.getKeyStroke(KeyEvent.VK_F2, 0))
                .name("Command")
                .tooltip("Execute a user-defined command")
                .icon(getIcon("terminal"))
        .get());

        actions.add(ab.init(getAction("shell", cs))
                .keyStroke(KeyStroke.getKeyStroke(KeyEvent.VK_F2, AppProperties.getInstance().getCtrlMask()))
                .name("Shell")
                .tooltip("Launch a command-line shell (FILES variable holds selected files)")
                .icon(getIcon("terminal"))
        .get());


        actions.add(ab.init(getAction("launch", cs))
                .keyStroke(KeyStroke.getKeyStroke(KeyEvent.VK_F4,0))
                .name("Launch")
                .tooltip("Launch selected file in OS default application")
                .icon(getIcon("exec"))
        .get());
       actions.add(ab.init(getAction("copy", cs))
                .keyStroke(KeyStroke.getKeyStroke(KeyEvent.VK_F5,0))
                .name("Copy")
                .tooltip("Copy selected file(s) to another directory")
                .icon(getIcon("editcopy"))
        .get());
        actions.add(ab.init(getAction("duplicate", cs))
                 .keyStroke(KeyStroke.getKeyStroke(KeyEvent.VK_F5,KeyEvent.SHIFT_MASK))
                 .name("Duplicate")
                 .tooltip("Duplicate the selected file")
                 .icon(getIcon("edit_add"))
         .get());
        actions.add(ab.init(getAction("move", cs))
                 .keyStroke(KeyStroke.getKeyStroke(KeyEvent.VK_F6,0))
                 .name("Move")
                 .tooltip("Move selected file(s) to another directory")
                 .icon(getIcon("filesaveas"))
         .get());
        actions.add(ab.init(getAction("rename", cs))
                .keyStroke(KeyStroke.getKeyStroke(KeyEvent.VK_F6,KeyEvent.SHIFT_MASK))
                .name("Rename")
                .tooltip("Rename the selected file")
                .icon(getIcon("filesaveas"))
                .get());
        actions.add(ab.init(getAction("mkdir", cs))
                .keyStroke(KeyStroke.getKeyStroke(KeyEvent.VK_F7,0))
                .name("Make dir")
                .tooltip("Create a new directory")
                .icon(getIcon("folder_new"))
                .get());
           actions.add(ab.init(getAction("delete", cs))
                .keyStroke(KeyStroke.getKeyStroke(KeyEvent.VK_F8,0))
                .name("Delete")
                .tooltip("Delete selected file(s)")
                .icon(getIcon("editdelete"))
                .get());
        return actions;
    }
    public JMenu createCommanderMenu(CommanduleSource cs){
        final JMenu commanderMenu = new JMenu("Commander");
        List<Action> actions= createCommanderActions(cs);
        for(Action a:actions) commanderMenu.add(a);
        return commanderMenu;
    }
    public JMenu createCommanderMenu(Commandule c){
        CommanduleSource cs = getWeakCommanduleSource(c);
        return createCommanderMenu(cs);
    }
    public List<Action> createCommanderActions(Commandule c){
        CommanduleSource cs = getWeakCommanduleSource(c);
        return createCommanderActions(cs);
    }

    private CommanduleSource getWeakCommanduleSource(Commandule c) {
        CommanduleSource cs = new CommanduleSource() {
            public Commandule getCommandule() {
                return actionsToCommandules.get(this);
            }
        };
        actionsToCommandules.put(cs,c);
        return cs;
    }

    final static Object[] args=new Object[0];
    final static Class[] argTypes=new Class[0];
    private Action getAction(final String methodName, final CommanduleSource cs){
        return new AbstractAction(){
            public void actionPerformed(ActionEvent e) {
//                if (log.isLoggable(Level.FINER)) {
//                    log.log(Level.FINER,"perform commandule action source="+e.getSource(),new Throwable());
//                }
                Object o = cs.getCommandule();//.get(this);
                if (o!=null) {
                    try {
                        final Method m = o.getClass().getMethod(methodName,argTypes);
                        try {
                            m.invoke(o,args);
                        } catch (IllegalAccessException e1) {
                            log.log(Level.WARNING, "exception", e1);
                        } catch (InvocationTargetException e1) {
                            log.log(Level.WARNING, "exception", e1);
                        }
                    } catch (NoSuchMethodException e1) {
                        log.log(Level.WARNING, "exception", e1);
                    }
                } else {
                    log.fine("commandule was null: ignoring action");
                }
            }
        };
    }
    private Icon getIcon(String name){
        return ImageProvider.getInstance().getIcon("icons/16/"+name+".png");
    }
    private Action setupAction(String iconName, Action copyAction, KeyStroke keystroke, String name,String tooltip){
        Action a= GuiUtilities.setupAction(copyAction,keystroke,name,name);
        a.putValue(Action.SMALL_ICON,getIcon(iconName));
         a.putValue(Action.SHORT_DESCRIPTION,tooltip);
        return a;
    }
}
