package uk.co.hcsoftware.commander.gui.util;

import javax.swing.*;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;



/**
 * operates a gui event buffer.
 * You specify the max buffer size (in units of your choice) and the minimum interval at which GUI
 * events will be posted on event dispatch thread
 *                    <br/>
 * Then call postEvent(..) with your updates. The method will block if there is no space in the buffer to store
 * the new event; otherwise it will be added to the queue.
 * Every so often SwingUtilities.invokeLater(..) will be called and all the runnables posted will be called in the
 * order they were posted
 *
 * Title: GuiUpdateQueue
 * Copyright:    Copyright (c) 04-Nov-2005
 *
 */
public class GuiUpdateQueue {
    public static final Logger log = Logger.getLogger(GuiUpdateQueue.class.getName());
    private Collection<Runnable> q;
    private final long guiUpdateInterval;
    private final long maxBuffer;
    private long bufferSize=0;
    private java.util.Timer t;
    private boolean taskScheduled=false;

    /**
     *
     * @param guiUpdateInterval rate at which events are posted on event queue, you will have to wait at least this long before
     * event is posted to EDT
     * @param bufferSize max size to buffer in input Q, when this size is reached/exceeded the input method
     * will block until the buffer size reduces. Units should be same as units passed to postEvent
     */
    public GuiUpdateQueue(long guiUpdateInterval, long bufferSize) {
       this(guiUpdateInterval,bufferSize,new LinkedList<Runnable>());
    }
    /**
     *
     * @param guiUpdateInterval rate at which events are posted on event queue, you will have to wait at least this long before
     * event is posted to EDT
     * @param bufferSize max size to buffer in input Q, when this size is reached/exceeded the input method
     * will block until the buffer size reduces. Units should be same as units passed to postEvent  use -1 for no limit
     * @param q Queue implementation (e.g. a LinkedList) this is exposed so can write own q (e.g. one that only stores the most recent runnable)
     */
    public GuiUpdateQueue(long guiUpdateInterval, long bufferSize, Collection<Runnable> q) {
        this.guiUpdateInterval = guiUpdateInterval;
        this.maxBuffer = bufferSize;
        t = new java.util.Timer();
        this.q=q;
    }


    /**
     * post a runnable that will be executed on ETD
     * @param r
     * @param eventSize  size requried on the heap to store this runnable. Same units as used in constructor buffersize method
     */
    public void postEvent(Runnable r, long eventSize){
        synchronized (q) {
            if(maxBuffer>0 && eventSize>=maxBuffer){
                //in this scenario, eventsize is bigger than the max buffer so execute as soon as have space
               while(bufferSize>0){
                    try {
                        if (log.isLoggable(Level.FINER))
                            log.finer("got event>maxbuffer, q is not empty: bufferSize="+bufferSize+" eventSize="+eventSize+" maxBuffer="+maxBuffer+" waiting..");
                        q.wait();
                    } catch (InterruptedException e) {
                        log.info("InterruptedException");
                    }
                }
            }
            else{
                while(maxBuffer>0 && (bufferSize+eventSize) > maxBuffer){
                    try {
                        if (log.isLoggable(Level.FINER))
                            log.finer("q is full! bufferSize="+bufferSize+" eventSize="+eventSize+" maxBuffer="+maxBuffer+" waiting..");
                        q.wait();
                    } catch (InterruptedException e) {
                        log.info("InterruptedException");
                    }
                }
            }
            if (log.isLoggable(Level.FINER))
                log.finer("add event to Q");
            q.add(r);
            bufferSize+=eventSize;
            //schedule dispatchEvents in guiUpdateInterval ms

            if (!taskScheduled) {
                if (log.isLoggable(Level.FINER))
                    log.finer("scheduling update in "+guiUpdateInterval+"ms");
                t.schedule(new TimerTask() {
                    public void run() {
                        synchronized (q) {
                            dispatchEvents();
                            taskScheduled=false;
                            q.notifyAll();
                        }
                    }
                },guiUpdateInterval);
                taskScheduled=true;
            }
        }
    }

    private void dispatchEvents(){
        synchronized (q) {
            if (log.isLoggable(Level.FINER))
                log.finer("dispatchEvents");
            if(q.isEmpty()){
                if (log.isLoggable(Level.FINER))
                    log.finer("dispatchEvents: nothing to dispatch!");
                return;
            }
            final List<Runnable> runnables = new ArrayList<Runnable>(q);
            SwingUtilities.invokeLater(new Runnable() {
                public void run() {
                    if (log.isLoggable(Level.FINER))
                        log.finer("executing "+runnables.size()+" runnables in EDT");
                    for(Runnable r:runnables) r.run();
                }
            });
            q.clear();
            bufferSize=0;
        }
    }
}
