package uk.co.hcsoftware.commander.gui.util;

import java.lang.ref.WeakReference;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.logging.Logger;


/**
 * The observer/observable pattern involves an observer class providing a reference to itself to some observable class.
 * This reference prevents the observable object being finalized and garbage-collected.
 * Often this is not what you want - if the observable is _only_ reachable by such a reference you probably want
 * it to be destroyed and the reference removed from the observer's list.
 * <br/>
 * This class provides a means of achieving this. it does the following:
 * <ol><li>stores the observer in a WeakReference</li><li>creates a proxy observer that forwards events to the real observer
 * if and only if it has not been finalized</li><li>if the object has been finalized, invokes onNullObserver(..) so you can unregister the listener</li></ol>
 *
 * <h2>Example:</h2><br/>
 * <pre><code>
 *  final JButton myButton..;
 *  final Action myAction..;
 *  WeakListener wl = new WeakListener(myAction, ActionListener.class){
 *     public Object onNullObserver() {
                myButton.removeActionlistener((Actionlistener) getProxy());
                return null;
       }
 * };
 *  myButton.addActionListener((Actionlistener)wl.getProxy());
 *
 *  <strong>NB:</strong> Be careful to use an observer class that won't get gc'd unless you want it to. E.g. don't use an anonymous inner
 * as this will be eligible for gc as soon as you leave the method! Use either the class that may be gc'd itself or make
 * the observer one of its private member variables.
 *
 *
 * Title: WeakListener
 * Copyright:    Copyright (c) 25-Nov-2005
 *
 * @author spurlinn
 */
public class WeakListener implements InvocationHandler {
    public static final Logger log = Logger.getLogger(WeakListener.class.getName());
    private Object proxy;

    public WeakListener(Object observer,Class observableInterface) {
        wr = new WeakReference(observer);

        //check observer instanceof if, illlegalarg if not
        if(!observableInterface.isAssignableFrom(observer.getClass())){
            throw new IllegalArgumentException("observer="+observer+" must be an instance of observableInterface="+observableInterface);
        }

        proxy = Proxy.newProxyInstance(
                observableInterface.getClassLoader()
                , new Class[]{observableInterface}
                , this
        );
    }

    private final WeakReference wr;

    public Object getProxy() {
        return proxy;
    }

    public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
        Object observable = wr.get();
        if(observable==null){
            log.finer("observable was null! not forwarding");
            //NB: maybe should unregister listener here
            return onNullObserver();
        }
        log.finest("forwarding to meth="+method.getName());
        //forward to the observable
        final Method meth = observable.getClass().getMethod(method.getName(),method.getParameterTypes());
        return meth.invoke(observable,args);
    }

    /**
     * override me (e.g. unregister the observable)
     * @return the object that will be returned to the Observer NB: make sure it is the correct type!
     */
    public Object onNullObserver(){
        return null;
    }
}
