package uk.co.hcsoftware.commander.gui.zip;

import uk.co.hcsoftware.commander.io.TempFileManager;

import java.io.*;
import java.util.*;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.zip.ZipOutputStream;


/**
 * Title: ZipNode
 * Copyright:    Copyright (c) 29-Nov-2005
 *
 * @author spurlinn
 */
public class ZipNode{
    private static final Logger log = Logger.getLogger(ZipNode.class.getName());
    private ZipEntry ze;
    private boolean isDirectory;
    private final ZipNode parent;
    private boolean syntheticNode;
    private File data;

    public ZipNode(ZipNode parent, String[] path, boolean directory) {
        this.parent=parent;
        this.path = path;
        isDirectory=directory;
        if(parent!=null) parent.addChild(this);
    }

    public static String[] getPath(ZipNode parent, String nodeName){
        String[] path=new String[parent.path.length+1];
        System.arraycopy(parent.path,0,path,0,parent.path.length);
        path[path.length-1]=nodeName;
        return path;
    }

// Here is a sample printout of ZipEntry.getName() from a well known zipfile (top-level dir is commanderbond-0.1)
//    commanderbond-0.1/src/uk/co/hcsoftware/commander/treeplugin/file/
//    commanderbond-0.1/src/uk/co/hcsoftware/commander/treeplugin/zip/
//    commanderbond-0.1/src/uk/co/hcsoftware/commander/vfs/
//    commanderbond-0.1/src/
//    commanderbond-0.1/src/com/
//    commanderbond-0.1/src/com/Ostermiller/
//    commanderbond-0.1/src/com/Ostermiller/util/
//    commanderbond-0.1/src/com/Ostermiller/util/Browser.java
//    commanderbond-0.1/src/com/Ostermiller/util/Browser.properties
//    commanderbond-0.1/src/com/Ostermiller/util/Browser_cz.properties
//    commanderbond-0.1/src/com/Ostermiller/util/Browser_da.properties
//    commanderbond-0.1/src/com/Ostermiller/util/Browser_de.properties
    /**
     * @return the path to this entry in the form used by ZipEntry.getName()
     */
    public String getZipPath(){
        StringBuilder sb = new StringBuilder();
        for(String elem:path){
            sb.append(elem);
            sb.append("/");
        }
        if(!isDirectory() && sb.length()>0){//strip off the trailing /
            sb.setLength(sb.length()-1);
        }
        return sb.toString();
    }

    private String[] getPath(String path) {
        final String[] strings = path.split("/");
        log.fine("path="+path+" yields="+Arrays.asList(strings));
        return strings;
    }

    public void setZipEntry(ZipEntry ze) {
        this.ze = ze;
        isDirectory=ze.isDirectory();
    }
    public void setSyntheticNode(InputStream is, boolean isDirectory) throws IOException {
        this.syntheticNode=true;
        this.isDirectory=(isDirectory); //if there is an inputstream then this must be a file
        if(!isDirectory){ //cache the data in the filesystem
//            File tempRoot=new File(FileUtils.tempDir(),"cache");
//            if(!tempRoot.exists()){
//                if (!tempRoot.mkdirs()) {
//                     throw new IOException("couldn't create temporary directory: "+tempRoot);
//                }
//            }
            File out = TempFileManager.createTempFile("synth-zipentry",".tmp");//File.createTempFile("synth-zipentry",".tmp",tempRoot);//new File(tempRoot,System.currentTimeMillis()+""+(uid++));
//            out.deleteOnExit();
            File parent = out.getParentFile();
            if(parent!=null){
                boolean b=parent.mkdirs();
                if(!b) log.warning("couldn't create parent of out="+out+" parent="+parent);
            }
            else log.fine("parent of out="+out+" was null!");
            OutputStream os = new BufferedOutputStream(new FileOutputStream(out));
            try {
                byte[] buf = new byte[1024];
                int i;
                while((i=is.read(buf))!=-1){
                    os.write(buf,0,i);
                }
            } finally {
                os.close();
            }
            log.fine("cached syntheticnode inputstream in file="+out);
            this.data=out;
        }
        log.fine("create synthetic node isdirectory="+this.isDirectory+" data="+data);
        ZipEntry ze = new ZipEntry(getZipPath());
        setZipEntry(ze);
    }

    private final String[] path; //full path to node, including name
    public String getName(){
        if(path.length==0) return "";
        return path[path.length-1];
    }
    public ZipNode getChild(String name){
        if(children==null) return null;
        for(ZipNode zn:children){
            if(name.equals(zn.getName())) return zn;
        }
        return null;
    }
    public ZipNode getOrCreateChild(String name){
        ZipNode zn = getChild(name);
        if(zn==null){
            String[] newPath = new String[path.length+1];
            System.arraycopy(path,0,newPath,0,path.length);
            newPath[newPath.length-1]=name;
            zn = new ZipNode(this,newPath, true);
        }
        return zn;
    }
    /**
     *
     * @param zipPath e.g. plugins/net.sourceforge.eclipseccase.source_1.1.0/src/net.sourceforge.eclipseccase.ui_1.1.0/icons/full/hijacked_ovr.gif
     * @return
     */
    public ZipNode getOrCreateDescendant(String zipPath) {
        String[] path = getPath(zipPath);
        ZipNode curr=this;
        for(String childName:path){
            curr=curr.getOrCreateChild(childName);
        }
        return curr;
    }
    public ZipNode getDescendant(String zipPath) {
        String[] path = getPath(zipPath);
        ZipNode curr=this;
        for(String childName:path){
            curr=curr.getChild(childName);
            if(curr==null) return null;
        }
        return curr;
    }


    private List<ZipNode> children;
    /**
     * adds this child to the parent zipNode's children
     *
     * WARNING: all children must have a unique name, so if the path of this child is identical to the path of any of the current children
     * then this node will replace the other child
     * @param zn
     */
    public void addChild(ZipNode zn){
        if(children==null) children=new ArrayList<ZipNode>();
        String zipPath=zn.getZipPath();
        //replae if duplicat path of any of the others
        for (ListIterator<ZipNode> iterator = children.listIterator(); iterator.hasNext();) {
            ZipNode zipNode = iterator.next();
            if(zipPath.equals(zipNode.getZipPath())){
                iterator.set(zn);
                return;
            }
        }
        children.add(zn);
    }
    public List<ZipNode> getChildren() {
        if(children==null) return Collections.emptyList();
        return children;
    }
    public boolean isDirectory() {
        return isDirectory;
    }
    public ZipNode getParent() {
        return parent;
    }

    public long getSize() {
        return ze==null?-1:ze.getSize();
    }

    public long getTime() {
        return ze==null?-1:ze.getTime();
    }
    public int getCompressedRatio() {
        if(ze==null) return -1;
        final long compressed_size = (ze.getCompressedSize()*100);
        final long original_size = (ze.getSize());
        //log.fine("comp="+compressed_size+" orig="+original_size);
        if(original_size==0 || original_size==-1) return -1;
        return 100-(int) (compressed_size/original_size);
    }

    public boolean hasEntry() {
        return ze!=null;
    }

    public ZipEntry getZipEntry() {
        return ze;
    }

    /**
     *
     * @return flat list of this node's zipentry (if it has one) followed by all its descendants' zipentries (if they have them)
     */
    public List<ZipEntry> getZipEntries() {
        LinkedList<ZipEntry> ll = new LinkedList<ZipEntry>();
        if(hasEntry()) ll.add(getZipEntry());
        if (children!=null) {
            for(ZipNode child:children){
                ll.addAll(child.getZipEntries());
            }
        }
        return ll;
    }
    /**
     *
     * @return flat list of this node's zipentry (if it has one) followed by all its descendants' zipentries (if they have them)
     */
     public List<ZipNode> getDescendants() {
        LinkedList<ZipNode> ll = new LinkedList<ZipNode>();
        ll.add(this);
        if (children!=null) {
            for(ZipNode child:children){
                ll.addAll(child.getDescendants());
            }
        }
        return ll;
    }

    /**
     * includes leading /
     * @return
     */
    private String getPathFromRoot() {
        StringBuffer sb = new StringBuffer();
        if(path.length==0) sb.append('/');
        for(String element:path){
            sb.append('/');
            sb.append(element);
        }
        return sb.toString();
    }
    /**
     * excludes leading /
     * @return
     */
    public String getPathRelativeToRoot(){
         return getPathFromRoot().substring(1);
    }

    public String toString() {
        return "name="+getName()+" path="+Arrays.asList(path)+" isdir="+isDirectory()+" synthetic?="+(isSyntheticNode());
    }

    public InputStream getInputStream(ZipFile zf) throws IOException {
        if(ze==null || ze.isDirectory()) return null;
        else if(isSyntheticNode()){
            log.finer("getInputstream getting data from datafile");
            if(data==null){
                log.warning("requested inputstream from synth node, but file was null! returning null");
                return null;
            }
            return new BufferedInputStream(new FileInputStream(data));
        }
        else return zf.getInputStream(ze);
    }                                       

    /**
     *
     * @return true if this node does not appear in teh original zipfile, but has been added later
     */
    public boolean isSyntheticNode() {
        return syntheticNode;//data!=null;
    }

    public void setDirectory(boolean directory) {
        isDirectory = directory;
    }

    public boolean removeFromParent() {
        ZipNode parent =getParent();
        if(parent!=null){
            return parent.children.remove(this);
        }
        else log.fine("failed to remove node from parent node="+this+" as parent was null");
        return false;
    }

    /**
     * copies this node to the new file, with everything the same but the name
     * @param newName
     * @param zf
     * @return
     * @throws CannotCreateEntryException
     * @throws IOException
     */
    public ZipNode cloneWithNewName(String newName, ZipFile zf) throws CannotCreateEntryException, IOException {
        if(parent==null){//can't rename the root!
            throw new CannotCreateEntryException("Cannot duplicate the root node!",newName);
        }
        ZipNode dupe = parent.getChild(newName);
        if(dupe!=null){//first check no collisions
            throw new CannotCreateEntryException("There is already a zip entry with that name",newName);
        }
        ZipNode zn = parent.getOrCreateChild(newName);
        zn.isDirectory=isDirectory;
        zn.syntheticNode=syntheticNode;
        if (!isDirectory) {
            InputStream is = getInputStream(zf);
            if(is!=null){
                ByteArrayOutputStream os = new ByteArrayOutputStream();
                byte[] buf = new byte[1024];
                int i;
                while((i=is.read(buf))!=-1){
                    os.write(buf,0,i);
                }
                is.close();
                zn.setSyntheticNode(new ByteArrayInputStream(os.toByteArray()), false);
            }
        }
        if(ze!=null){
            try {
                ZipEntry ze1 = new ZipEntry(zn.getZipPath());
                final long time = ze.getTime();
                if(time!=-1) ze1.setTime(time);
//            ze1.setCrc(ze.getCrc());
//                final long size = ze.getSize();
//                if(size>=0) ze1.setSize(size);
//                final int method = ze.getMethod();
//                if(method!=-1) ze1.setMethod(method);
                final byte[] extra = ze.getExtra();
                if(extra!=null) ze1.setExtra(extra);
                final String comment = ze.getComment();
                if(comment!=null) ze1.setComment(comment);
                zn.setZipEntry(ze1);
            } catch (Exception e) {
                log.log(Level.FINE, "couldn't set zipentry - skipping", e);
            }
        }
        return zn;
    }

    /**
     * write this zipentry to the outputstream, fetch data from zf as necessary
     * @param zf
     * @param zos
     * @throws IOException
     */
    public void write(ZipFile zf, ZipOutputStream zos) throws IOException {
        if(parent==null) return; //we don't write the root node
        zos.setMethod(ZipOutputStream.DEFLATED);
        final ZipEntry ze;
        if(hasEntry()){
//            final int method = this.ze.getMethod();
//            if (method== ZipOutputStream.DEFLATED || method== ZipOutputStream.STORED) {
//                zos.setMethod(method);
//            }
            ze=this.ze;
        }
        else {
           ze=new ZipEntry(getZipPath());
        }
        zos.putNextEntry(ze);
        InputStream is = getInputStream(zf);
        if (is!=null) {
            int i;
            byte[] buf=new byte[1024];
            while ((i = is.read(buf)) !=-1) {
                zos.write(buf, 0, i);
            }
            is.close();
        }
    }
}
