package uk.co.hcsoftware.commander.io;

import ns.PushbackIterator;
import ns.TaskTimer;

import java.util.logging.Logger;
import java.util.logging.Level;
import java.util.*;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.io.File;


/**
 * Title: DirectoryWatcher
 * Copyright:    Copyright (c) 25-Apr-2006
 *
 * @author N Spurling
 *
 * //listen for dir changes and fire filemodel events if the files in this dir change
 * 1. register for frameactivation events with commanderframe (via weaklistener)
 * 2. on activ - check if dircontents changed (using File.list() file.getName() - don't use lastModified as is native)
 * 3. if changed fire event to say what's been added/removed
 * 4. have filetablemodel create one of these per dir
 */
public abstract class DirectoryWatcher implements Runnable{
    public static final Logger log = Logger.getLogger(DirectoryWatcher.class.getName());
    private final File dir;
    SortedSet<String> cached;
    private Future future;

    public DirectoryWatcher(File dir) {
        this.dir = dir;
        if(dir.exists()){
            String[] names = dir.list();
            cached=new TreeSet<String>();
            for(String name:names) cached.add(name);
        }
    }
    TaskTimer tt;
    String result;
    private static final ExecutorService es = Executors.newFixedThreadPool(2);
    //threading works as follows:
    //all instances of this class serialize their directory checks on a shared threadpool (to avoid too many threads, since we have one of these classes per dir-view)
    //we use 2 threads to try and avoid getting bogged down where some dirs are slow and some are fast..
    //If any one instance gets another checkForChanges() before it's prev checkForChanges has started, cancels prev
    public void checkForChanges(){
        // run in separate thread
        synchronized (this) {
            if(future!=null) future.cancel(false);
            future = es.submit(this);
        }
    }

    public void run() {
        if(log.isLoggable(Level.FINE)){
            tt=new TaskTimer();
            result="";
        }
        List<String> current = Arrays.asList(dir.list());
        if(log.isLoggable(Level.FINE)){
            result+=tt.timeTakenSoFar()+" to list "+dir;
        }
        Collections.sort(current);

        //everything in current that is not in cached has been added
        List<String> added = removeAll(current,cached);
        //everything in cached that is not in current has been deleted
        List<String> deleted = removeAll(cached,current);
        cached.addAll(added);
        cached.removeAll(deleted);
        if(log.isLoggable(Level.FINE)){
            log.fine(result+" "+tt.stopString()+" total");
        }
        if(!added.isEmpty() || !deleted.isEmpty()){
            log.fine("dir="+dir+" added="+added.size()+" del="+deleted.size());
            changed(added,deleted);
        }
        else log.fine("dir="+dir+"not firing changes, as nothing changed");
    }

    public abstract void changed(List<String> added, List<String> deleted) ;

    /**
     * remove from c1 everything that is in the c2 - faster than Collections.removeAll
     * but all objects must be comparables
     * @param c1 must be sorted
     * @param c2 must be sorted
     * @return list of only those items in c1 that are not in c2
     */
    private List<String> removeAll(Collection<String> c1, Collection<String> c2){
        //first sort c1 and c2
        List<String> al = new ArrayList<String>();
        PushbackIterator<String> i = new PushbackIterator<String>(c1.iterator());
        Iterator<String> j = c2.iterator();
        if(!i.hasNext()) return Collections.emptyList();
        if(!j.hasNext()) return new ArrayList<String>(c1);
        String o1=null;
        while(j.hasNext()){
            o1 = j.next();
            //add all before this comparable2
            while (i.hasNext()) {
                String c= i.next();
                int cv = c.compareTo(o1);
                if(cv<0) al.add(c);
                else{
                    if(cv>0) i.push(c);
                    break;
                }
            }
        }
        //add all after the last comparable2
        while (i.hasNext()) {
            String c = i.next();
            int cv = c.compareTo(o1);
            if(cv>0){
                al.add(c);
            }
        }

        return al;
    }
}
