package uk.co.hcsoftware.commander.io;

import uk.co.hcsoftware.commander.io.treestream.Node;
import uk.co.hcsoftware.commander.io.treestream.TreeStream;
import uk.co.hcsoftware.commander.io.treestream.DefaultNode;

import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.util.List;
import java.util.Collections;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;

/**
 * similar to FileListTransferable but finds the common ancestor to all the files in the list
 * and provides a directory structure that preserves these relationships
 *
 * @author N Spurling
 */
public class StructuredFileListTransferable implements Transferable {
    public static final Logger log = Logger.getLogger(StructuredFileListTransferable.class.getName());

    public static final DataFlavor[] FLAVORS={TreeStream.DATAFLAVOR};
    private TreeStream ts;
    public StructuredFileListTransferable(List<File> files, Object source) throws CannotFindCommonAncestorException, IOException {
        final File root;
        try {
            root = FileUtils.getInstance().findCommonAncestor(files);
            if(root==null) throw new CannotFindCommonAncestorException(files);
        } catch (IOException e) {
            log.log(Level.WARNING,"exception",e);
            throw new CannotFindCommonAncestorException(files);
        }
        DefaultNode rootNode = new DefaultNode(root.getName());
        //now for each file put it in the correct place
        String rootPath=root.getCanonicalPath();
        for(File f:files){
            String pathFromRoot = f.getCanonicalPath().substring(rootPath.length());
            //it is possible that parent is the 
            log.finer("root="+rootPath+" path="+pathFromRoot);
            //strip off the pathname
            String pathOnly = pathFromRoot.substring(0,pathFromRoot.length()-f.getName().length());
            log.finer("path="+pathOnly);
            //strip off any leading/trailing pathseps
            if(pathOnly.startsWith(File.separator)) pathOnly=pathOnly.substring(1);
            if(pathOnly.endsWith(File.separator)) pathOnly=pathOnly.substring(0,pathOnly.length()-1);
            log.finer("path="+pathOnly);
            //convert to array
            String[] path = FileUtils.getInstance().splitPath(pathOnly);
            //starting at root, traverse tree
            DefaultNode node = rootNode;
            for(String dir:path){
                Node child = node.getChild(dir);
                if(child!=null && !(child instanceof DefaultNode)){
                    log.warning("path="+path+" dir="+dir+" files="+files);
                    throw new IOException("Invalid directory structure!");
                }
                DefaultNode n = (DefaultNode) child;
                if(n==null){
                    n = new DefaultNode(dir);
                    node.addChild(n);
                }
                node=n;
            }
            //now add filenode leaf
            node.addChild(new FileNode(f));
        }


        List<? extends Node> roots= Collections.singletonList(rootNode);
        ts=new TreeStream(roots,source);

    }

    public DataFlavor[] getTransferDataFlavors() {
        return StructuredFileListTransferable.FLAVORS;
    }
    public boolean isDataFlavorSupported(DataFlavor flavor) {
        for(DataFlavor f: StructuredFileListTransferable.FLAVORS) if(f.match(flavor)) return true;
        return false;
    }
    List<File> dragFilesList ;
    public Object getTransferData(DataFlavor flavor) throws UnsupportedFlavorException {
        if(TreeStream.DATAFLAVOR.match(flavor)){
            return ts;
        }
        else throw new UnsupportedFlavorException(flavor);
    }
}
